# Copyright (c) 2020 PyI40AAS Contributors
#
# This program and the accompanying materials are made available under the terms of the Eclipse Public License v. 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0 which is available
# at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
"""
Module which offers functions to use in a confirmation tool related to AASX files

check_deserialization: Checks if a AASX file can be deserialized

check_aas_example: Checks if a AASX file consist the data of the example data defined in
                   aas.examples.data.example_aas.py

check_aasx_files_equivalence: Checks if two AASX files have the same data regardless of their order

All functions reports any issues using the given StateManager by adding new steps and associated LogRecords
"""
import datetime
import json
import logging
from typing import Optional, Tuple

import pyecma376_2

from .. import model
from ..adapter import aasx
from ..examples.data import example_aas, create_example_aas_binding
from ..examples.data._helper import AASDataChecker, DataChecker
from .state_manager import ComplianceToolStateManager, Status


def check_deserialization(file_path: str, state_manager: ComplianceToolStateManager,
                          file_info: Optional[str] = None) \
        -> Tuple[model.DictObjectStore, aasx.DictSupplementaryFileContainer, pyecma376_2.OPCCoreProperties]:
    """
    Read a AASX file and reports any issues using the given StateManager

    add the steps: 'Open {} file' and 'Read {} file'

    :param file_path: given file which should be deserialized
    :param state_manager: manager to log the steps
    :param file_info: additional information about the file for name of the steps
    :return: returns the read object store
    """
    logger = logging.getLogger('compliance_check')
    logger.addHandler(state_manager)
    logger.propagate = False
    logger.setLevel(logging.INFO)

    # create handler to get logger info
    logger_deserialization = logging.getLogger(aasx.__name__)
    logger_deserialization.addHandler(state_manager)
    logger_deserialization.propagate = False
    logger_deserialization.setLevel(logging.INFO)

    if file_info:
        state_manager.add_step('Open {} file'.format(file_info))
    else:
        state_manager.add_step('Open file')
    try:
        # open given file
        reader = aasx.AASXReader(file_path)
        state_manager.set_step_status_from_log()
    except ValueError as error:
        logger.error(error)
        state_manager.set_step_status_from_log()
        state_manager.add_step('Read file')
        state_manager.set_step_status(Status.NOT_EXECUTED)
        return model.DictObjectStore(), aasx.DictSupplementaryFileContainer(), pyecma376_2.OPCCoreProperties()

    try:
        # read given file
        state_manager.add_step('Read file')
        obj_store: model.DictObjectStore[model.Identifiable] = model.DictObjectStore()
        files = aasx.DictSupplementaryFileContainer()
        reader.read_into(obj_store, files)
        new_cp = reader.get_core_properties()
        state_manager.set_step_status(Status.SUCCESS)
    except ValueError as error:
        logger.error(error)
        state_manager.set_step_status(Status.FAILED)
        return model.DictObjectStore(), aasx.DictSupplementaryFileContainer(), pyecma376_2.OPCCoreProperties()
    finally:
        reader.close()

    return obj_store, files, new_cp


def check_aas_example(file_path: str, state_manager: ComplianceToolStateManager) -> None:
    """
    Checks if a file contains all elements of the aas example and reports any issues using the given StateManager

    calls the check_deserialization and add the steps: 'Check if data is equal to example data'

    :param file_path: given file which should be checked
    :param state_manager: manager to log the steps
    """
    logger = logging.getLogger('compliance_check')
    logger.addHandler(state_manager)
    logger.propagate = False
    logger.setLevel(logging.INFO)

    # create handler to get logger info
    logger_example = logging.getLogger(example_aas.__name__)
    logger_example.addHandler(state_manager)
    logger_example.propagate = False
    logger_example.setLevel(logging.INFO)

    obj_store, files, cp_new = check_deserialization(file_path, state_manager)

    if state_manager.status in (Status.FAILED, Status.NOT_EXECUTED):
        state_manager.add_step('Check if data is equal to example data')
        state_manager.set_step_status(Status.NOT_EXECUTED)
        state_manager.add_step('Check if core properties are equal')
        state_manager.set_step_status(Status.NOT_EXECUTED)
        return

    checker = AASDataChecker(raise_immediately=False)

    state_manager.add_step('Check if data is equal to example data')
    example_data = create_example_aas_binding()
    checker.check_object_store(obj_store, example_data)
    state_manager.add_log_records_from_data_checker(checker)

    if state_manager.status in (Status.FAILED, Status.NOT_EXECUTED):
        state_manager.add_step('Check if core properties are equal')
        state_manager.set_step_status(Status.NOT_EXECUTED)
        return

    state_manager.add_step('Check if core properties are equal')
    # Create OPC/AASX core properties
    cp = pyecma376_2.OPCCoreProperties()
    cp.created = datetime.datetime(2020, 1, 1, 0, 0, 0)
    cp.creator = "PyI40AAS Testing Framework"
    cp.description = "Test_Description"
    cp.lastModifiedBy = "PyI40AAS Testing Framework Compliance Tool"
    cp.modified = datetime.datetime(2020, 1, 1, 0, 0, 1)
    cp.revision = "1.0"
    cp.version = "2.0.1"
    cp.title = "Test Title"

    checker2 = DataChecker(raise_immediately=False)
    assert (isinstance(cp_new.created, datetime.datetime))
    duration = cp_new.created - cp.created
    checker2.check(duration.microseconds < 20, "created must be {}".format(cp.created), created=cp_new.created)
    checker2.check(cp_new.creator == cp.creator, "creator must be {}".format(cp.creator), creator=cp_new.creator)
    checker2.check(cp_new.description == cp.description, "description must be {}".format(cp.description),
                   description=cp_new.description)
    checker2.check(cp_new.lastModifiedBy == cp.lastModifiedBy, "lastModifiedBy must be {}".format(cp.lastModifiedBy),
                   lastModifiedBy=cp_new.lastModifiedBy)
    assert (isinstance(cp_new.modified, datetime.datetime))
    duration = cp_new.modified - cp.modified
    checker2.check(duration.microseconds < 20, "modified must be {}".format(cp.modified), modified=cp_new.modified)
    checker2.check(cp_new.revision == cp.revision, "revision must be {}".format(cp.revision), revision=cp_new.revision)
    checker2.check(cp_new.version == cp.version, "version must be {}".format(cp.version), version=cp_new.version)
    checker2.check(cp_new.title == cp.title, "title must be {}".format(cp.title), title=cp_new.title)

    # Check if file in file object is the same
    list_of_id_shorts = ["ExampleSubmodelCollectionUnordered", "ExampleFile"]
    obj = example_data.get_identifiable(model.Identifier("https://acplt.org/Test_Submodel", model.IdentifierType.IRI))
    for id_short in list_of_id_shorts:
        obj = obj.get_referable(id_short)
    obj2 = obj_store.get_identifiable(model.Identifier("https://acplt.org/Test_Submodel", model.IdentifierType.IRI))
    for id_short in list_of_id_shorts:
        obj2 = obj2.get_referable(id_short)
    try:
        sha_file = files.get_sha256(obj.value)
    except KeyError as error:
        state_manager.add_log_records_from_data_checker(checker2)
        logger.error(error)
        state_manager.set_step_status(Status.FAILED)
        return

    checker2.check(sha_file == files.get_sha256(obj2.value), "File of {} must be {}.".format(obj.value, obj2.value),
                   value=obj2.value)
    state_manager.add_log_records_from_data_checker(checker2)
    if state_manager.status in (Status.FAILED, Status.NOT_EXECUTED):
        state_manager.set_step_status(Status.FAILED)
    else:
        state_manager.set_step_status(Status.SUCCESS)


def check_aasx_files_equivalence(file_path_1: str, file_path_2: str, state_manager: ComplianceToolStateManager) -> None:
    """
    Checks if two aasx files contain the same elements in any order and reports any issues using the given StateManager

    calls the check_deserialization for ech file and add the steps: 'Check if data in files are equal'

    :param file_path_1: given first file which should be checked
    :param file_path_2: given second file which should be checked
    :param state_manager: manager to log the steps
    """
    logger = logging.getLogger('compliance_check')
    logger.addHandler(state_manager)
    logger.propagate = False
    logger.setLevel(logging.INFO)

    obj_store_1, files_1, cp_1 = check_deserialization(file_path_1, state_manager, 'first')

    obj_store_2, files_2, cp_2 = check_deserialization(file_path_2, state_manager, 'second')

    if state_manager.status is Status.FAILED:
        state_manager.add_step('Check if data in files are equal')
        state_manager.set_step_status(Status.NOT_EXECUTED)
        state_manager.add_step('Check if core properties are equal')
        state_manager.set_step_status(Status.NOT_EXECUTED)
        return

    checker = AASDataChecker(raise_immediately=False)
    try:
        state_manager.add_step('Check if data in files are equal')
        checker.check_object_store(obj_store_1, obj_store_2)
    except (KeyError, AssertionError) as error:
        state_manager.set_step_status(Status.FAILED)
        logger.error(error)
        state_manager.add_step('Check if core properties are equal')
        state_manager.set_step_status(Status.NOT_EXECUTED)
        return

    state_manager.add_log_records_from_data_checker(checker)

    if state_manager.status is Status.FAILED:
        state_manager.add_step('Check if core properties are equal')
        state_manager.set_step_status(Status.NOT_EXECUTED)
        return

    state_manager.add_step('Check if core properties are equal')
    checker2 = DataChecker(raise_immediately=False)
    assert (isinstance(cp_1.created, datetime.datetime))
    assert (isinstance(cp_2.created, datetime.datetime))
    duration = cp_1.created - cp_2.created
    checker2.check(duration.microseconds < 20, "created must be {}".format(cp_1.created), value=cp_2.created)
    checker2.check(cp_1.creator == cp_2.creator, "creator must be {}".format(cp_1.creator), value=cp_2.creator)
    checker2.check(cp_1.lastModifiedBy == cp_2.lastModifiedBy, "lastModifiedBy must be {}".format(cp_1.lastModifiedBy),
                   value=cp_2.lastModifiedBy)
    state_manager.add_log_records_from_data_checker(checker2)



#!/usr/bin/env python3
# Copyright (c) 2019-2021 PyI40AAS Contributors
#
# This program and the accompanying materials are made available under the terms of the Eclipse Public License v. 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0 which is available
# at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0

import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyi40aas",
    version="0.2.2",
    author="Chair of Process Control Engineering, RWTH Aachen",
    author_email="m.thies@plt.rwth-aachen.de",
    description="An implementation of the Asset Administration Shell for Industry 4.0 systems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.rwth-aachen.de/acplt/pyi40aas",
    packages=setuptools.find_packages(exclude=["test", "test.*"]),
    zip_safe=False,
    package_data={
        "aas": ["py.typed"],
        "aas.adapter.json": ["aasJSONSchema.json"],
        "aas.adapter.xml": ["AAS.xsd", "AAS_ABAC.xsd", "IEC61360.xsd"],
        "aas.examples.data": ["TestFile.pdf"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    entry_points={
        'console_scripts': [
            "aas_compliance_check = aas.compliance_tool.cli:main"
        ]
    },
    python_requires='>=3.6',
    install_requires=[
        'python-dateutil>=2.8,<3',
        'lxml>=4.2,<5',
        'urllib3>=1.26<2.0',
        'pyecma376-2>=0.2.4',
    ]
)



# Copyright (c) 2020 PyI40AAS Contributors
#
# This program and the accompanying materials are made available under the terms of the Eclipse Public License v. 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0 which is available
# at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
"""
This module contains the classes `ConceptDescription` and `ConceptDictionary` from the AAS meta model as well as
specialized ConceptDescriptions like `IEC61360ConceptDescription`.
"""
from enum import unique, Enum
from typing import Optional, Set, Type

from . import base, datatypes


class ConceptDescription(base.Identifiable):
    """
    The semantics of a property or other elements that may have a semantic description is defined by a concept
    description.

    The description of the concept should follow a standardized schema (realized as data specification template).

    :ivar is_case_of: Unordered list of global references to external definitions the concept is compatible to or was
                      derived from.
                      Note: Compare to is-case-of relationship in ISO 13584-32 & IEC EN 61360
    """

    def __init__(self,
                 identification: base.Identifier,
                 is_case_of: Optional[Set[base.Reference]] = None,
                 id_short: str = "",
                 category: Optional[str] = None,
                 description: Optional[base.LangStringSet] = None,
                 parent: Optional[base.Namespace] = None,
                 administration: Optional[base.AdministrativeInformation] = None):
        """
        Initializer of ConceptDescription

        :param identification: The globally unique identification of the element. (from base.Identifiable)
        :param is_case_of: Unordered list of global references to external definitions the concept is compatible to or
                           was derived from.
                           Note: Compare to is-case-of relationship in ISO 13584-32 & IEC EN 61360
        :param id_short: Identifying string of the element within its name space. (from base.Referable)
        :param category: The category is a value that gives further meta information w.r.t. to the class of the element.
                         It affects the expected existence of attributes and the applicability of constraints.
                         (from base.Referable)
        :param description: Description or comments on the element. (from base.Referable)
        :param parent: Reference to the next referable parent element of the element. (from base.Referable)
        :param administration: Administrative information of an identifiable element. (from base.Identifiable)
        """
        super().__init__()
        self.identification: base.Identifier = identification
        self.is_case_of: Set[base.Reference] = set() if is_case_of is None else is_case_of
        self.id_short = id_short
        self.category: Optional[str] = category
        self.description: Optional[base.LangStringSet] = dict() if description is None else description
        self.parent: Optional[base.Namespace] = parent
        self.administration: Optional[base.AdministrativeInformation] = administration


class ConceptDictionary(base.Referable):
    """
    A dictionary containing concept descriptions.

    Typically a concept description dictionary of an AAS contains only concept descriptions of elements used within
    submodels of the AAS.


    :param concept_description: Unordered list of references to elements of class ConceptDescription
    """
    def __init__(self,
                 id_short: str,
                 category: Optional[str] = None,
                 description: Optional[base.LangStringSet] = None,
                 parent: Optional[base.Namespace] = None,
                 concept_description: Optional[Set[base.AASReference[ConceptDescription]]] = None):
        """
        Initializer of ConceptDictionary

        :param id_short: Identifying string of the element within its name space. (from base.Referable)
        :param category: The category is a value that gives further meta information w.r.t. to the class of the element.
                         It affects the expected existence of attributes and the applicability of constraints. (from
                         base.Referable)
        :param description: Description or comments on the element. (from base.Referable)
        :param parent: Reference to the next referable parent element of the element. (from base.Referable)
        :param concept_description: Unordered list of references to elements of class ConceptDescription

        TODO: Add instruction what to do after construction
        """
        super().__init__()
        self.id_short = id_short
        self.category: Optional[str] = category
        self.description: Optional[base.LangStringSet] = dict() if description is None else description
        self.parent: Optional[base.Namespace] = parent
        self.concept_description: Set[base.AASReference[ConceptDescription]] = \
            set() if concept_description is None else concept_description


# #############################################################################
# Helper types for
@unique
class IEC61360DataType(Enum):
    """
    Data types for data_type in DataSpecificationIEC61360
    """
    DATE = 0
    STRING = 1
    STRING_TRANSLATABLE = 2
    REAL_MEASURE = 3
    REAL_COUNT = 4
    REAL_CURRENCY = 5
    BOOLEAN = 6
    URL = 7
    RATIONAL = 8
    RATIONAL_MEASURE = 9
    TIME = 10
    TIMESTAMP = 11


@unique
class IEC61360LevelType(Enum):
    """
    Level types for the level_type in DataSpecificationIEC61360
    """
    MIN = 0
    MAX = 1
    NOM = 2
    TYP = 3


class IEC61360ConceptDescription(ConceptDescription):
    """
    A specialized ConceptDescription to define concepts according to IEC61360
    """
    def __init__(self,
                 identification: base.Identifier,
                 preferred_name: base.LangStringSet,
                 data_type: Optional[IEC61360DataType] = None,
                 definition: Optional[base.LangStringSet] = None,
                 short_name: Optional[base.LangStringSet] = None,
                 is_case_of: Optional[Set[base.Reference]] = None,
                 id_short: str = "",
                 category: Optional[str] = None,
                 description: Optional[base.LangStringSet] = None,
                 parent: Optional[base.Namespace] = None,
                 administration: base.AdministrativeInformation = None,
                 unit: Optional[str] = None,
                 unit_id: Optional[base.Reference] = None,
                 source_of_definition: Optional[str] = None,
                 symbol: Optional[str] = None,
                 value_format: base.DataTypeDef = None,
                 value_list: Optional[base.ValueList] = None,
                 value: Optional[base.ValueDataType] = None,
                 value_id: Optional[base.Reference] = None,
                 level_types: Set[IEC61360LevelType] = None,
                 ):
        """
        Initializer of IEC61360ConceptDescription

        :param identification: The globally unique identification of the element. (from base.Identifiable)
        :param preferred_name: preferred of the data object
        :param short_name: short name of the data object
        :param data_type: data type of the data object
        :param definition: definition of the data object
        :param is_case_of: Unordered list of global references to external definitions the concept is compatible to or
                           was derived from.
                           Note: Compare to is-case-of relationship in ISO 13584-32 & IEC EN 61360
        :param id_short: Identifying string of the element within its name space. (from base.Referable)
        :param category: The category is a value that gives further meta information w.r.t. to the class of the element.
                         It affects the expected existence of attributes and the applicability of constraints. (from
                         base.Referable)
        :param description: Description or comments on the element. (from base.Referable)
        :param parent: Reference to the next referable parent element of the element. (from base.Referable)
        :param administration: Administrative information of an identifiable element. (from base.Identifiable)
        :param unit: unit of the data object (optional)
        :param unit_id: reference to a unit id (optional)
        :param source_of_definition: source of the definition (optional)
        :param symbol: unit symbol (optional)
        :param value_format: format of the values (optional)
        :param value_list: list of values (optional)
        :param value: value data type object (optional)
        :param value_id: Reference to the value (optional)
        :param level_types: Set of level types of the DataSpecificationContent (optional)
        """
        super().__init__(identification, is_case_of, id_short, category, description, parent, administration)
        self.preferred_name: base.LangStringSet = preferred_name
        self.short_name: Optional[base.LangStringSet] = short_name
        self.data_type: Optional[IEC61360DataType] = data_type
        self.definition: Optional[base.LangStringSet] = definition
        self.unit: Optional[str] = unit
        self.unit_id: Optional[base.Reference] = unit_id
        self.source_of_definition: Optional[str] = source_of_definition
        self.symbol: Optional[str] = symbol
        self.value_list: Optional[base.ValueList] = value_list
        self.value_id: Optional[base.Reference] = value_id
        self.level_types: Set[IEC61360LevelType] = level_types if level_types else set()
        self.value_format: Optional[Type[datatypes.AnyXSDType]] = value_format
        self._value: Optional[base.ValueDataType] = (datatypes.trivial_cast(value, self.value_format)
                                                     if (value is not None and self.value_format is not None) else None)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value) -> None:
        if value is None or self.value_format is None:
            self._value = None
        else:
            self._value = datatypes.trivial_cast(value, self.value_format)



"""
This package contains functionality for serialization and deserialization of PyI40AAS objects into/from XML.

xml_serialization:
    The module offers a function to write an ObjectStore to a given file.

xml_deserialization.py
    The module offers a function to create an ObjectStore from a given xml document.
"""
import os.path

from .xml_serialization import write_aas_xml_file
from .xml_deserialization import AASFromXmlDecoder, StrictAASFromXmlDecoder, StrippedAASFromXmlDecoder, \
    StrictStrippedAASFromXmlDecoder, XMLConstructables, read_aas_xml_file, read_aas_xml_file_into, read_aas_xml_element

XML_SCHEMA_FILE = os.path.join(os.path.dirname(__file__), 'AAS.xsd')



# Copyright (c) 2020 PyI40AAS Contributors
#
# This program and the accompanying materials are made available under the terms of the Eclipse Public License v. 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0 which is available
# at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
"""
Module for serializing Asset Administration Shell data to the official XML format

How to use:
- For generating an XML-File from a model.registry.AbstractObjectStore, check out the function "write_aas_xml_file".
- For serializing any object to an XML fragment, that fits the XML specification from 'Details of the
  Asset Administration Shell', chapter 5.4, check out `<your_object_class_name_here>_to_xml()`. These functions return
  an xml.etree.ElementTree.Element object to be serialized into XML.
"""

from lxml import etree  # type: ignore
from typing import Dict, IO, Optional
import base64

from aas import model
from .. import _generic


# ##############################################################
# functions to manipulate etree.Elements more effectively
# ##############################################################

# Namespace definition
NS_AAS = "{http://www.admin-shell.io/aas/2/0}"
NS_ABAC = "{http://www.admin-shell.io/aas/abac/2/0}"
NS_AAS_COMMON = "{http://www.admin-shell.io/aas_common/2/0}"
NS_XSI = "{http://www.w3.org/2001/XMLSchema-instance}"
NS_XS = "{http://www.w3.org/2001/XMLSchema}"
NS_IEC = "{http://www.admin-shell.io/IEC61360/2/0}"
NS_MAP = {"aas": "http://www.admin-shell.io/aas/2/0",
          "abac": "http://www.admin-shell.io/aas/abac/2/0",
          "aas_common": "http://www.admin-shell.io/aas_common/2/0",
          "xsi": "http://www.w3.org/2001/XMLSchema-instance",
          "IEC": "http://www.admin-shell.io/IEC61360/2/0",
          "xs": "http://www.w3.org/2001/XMLSchema"}


def _generate_element(name: str,
                      text: Optional[str] = None,
                      attributes: Optional[Dict] = None) -> etree.Element:
    """
    generate an ElementTree.Element object

    :param name: namespace+tag_name of the element
    :param text: Text of the element. Default is None
    :param attributes: Attributes of the elements in form of a dict {"attribute_name": "attribute_content"}
    :return: ElementTree.Element object
    """
    et_element = etree.Element(name)
    if text:
        et_element.text = text
    if attributes:
        for key, value in attributes.items():
            et_element.set(key, value)
    return et_element


def boolean_to_xml(obj: bool) -> str:
    """
    serialize a boolean to XML

    :param obj: boolean
    :return: string in the XML accepted form
    """
    if obj:
        return "true"
    else:
        return "false"


# ##############################################################
# transformation functions to serialize abstract classes from model.base
# ##############################################################


def abstract_classes_to_xml(tag: str, obj: object) -> etree.Element:
    """
    Generates an XML element and adds attributes of abstract base classes of `obj`.

    If the object obj is inheriting from any abstract AAS class, this function adds all the serialized information of
    those abstract classes to the generated element.

    :param tag: tag of the element
    :param obj: an object of the AAS
    :return: parent element with the serialized information from the abstract classes
    """
    elm = _generate_element(tag)
    if isinstance(obj, model.Referable):
        elm.append(_generate_element(name=NS_AAS + "idShort", text=obj.id_short))
        if obj.category:
            elm.append(_generate_element(name=NS_AAS + "category", text=obj.category))
        if obj.description:
            elm.append(lang_string_set_to_xml(obj.description, tag=NS_AAS + "description"))
    if isinstance(obj, model.Identifiable):
        elm.append(_generate_element(name=NS_AAS + "identification",
                                     text=obj.identification.id,
                                     attributes={"idType": _generic.IDENTIFIER_TYPES[obj.identification.id_type]}))
        if obj.administration:
            elm.append(administrative_information_to_xml(obj.administration))
    if isinstance(obj, model.HasKind):
        if obj.kind is model.ModelingKind.TEMPLATE:
            elm.append(_generate_element(name=NS_AAS + "kind", text="Template"))
        else:
            # then modeling-kind is Instance
            elm.append(_generate_element(name=NS_AAS + "kind", text="Instance"))
    if isinstance(obj, model.HasSemantics):
        if obj.semantic_id:
            elm.append(reference_to_xml(obj.semantic_id, tag=NS_AAS+"semanticId"))
    if isinstance(obj, model.Qualifiable):
        if obj.qualifier:
            for qualifier in obj.qualifier:
                et_qualifier = _generate_element(NS_AAS+"qualifier")
                if isinstance(qualifier, model.Qualifier):
                    et_qualifier.append(qualifier_to_xml(qualifier, tag=NS_AAS+"qualifier"))
                if isinstance(qualifier, model.Formula):
                    et_qualifier.append(formula_to_xml(qualifier, tag=NS_AAS+"formula"))
                elm.append(et_qualifier)
    return elm


# ##############################################################
# transformation functions to serialize classes from model.base
# ##############################################################


def _value_to_xml(value: model.ValueDataType,
                  value_type: model.DataTypeDef,
                  tag: str = NS_AAS+"value") -> etree.Element:
    """
    Serialization of objects of class ValueDataType to XML

    :param value: model.ValueDataType object
    :param value_type: Corresponding model.DataTypeDef
    :param tag: tag of the serialized ValueDataType object
    :return: Serialized ElementTree.Element object
    """
    # todo: add "{NS_XSI+"type": "xs:"+model.datatypes.XSD_TYPE_NAMES[value_type]}" as attribute, if the schema allows
    #  it
    return _generate_element(tag,
                             text=model.datatypes.xsd_repr(value))


def lang_string_set_to_xml(obj: model.LangStringSet, tag: str) -> etree.Element:
    """
    serialization of objects of class LangStringSet to XML

    :param obj: object of class LangStringSet
    :param tag: tag name of the returned XML element (incl. namespace)
    :return: serialized ElementTree object
    """
    et_lss = _generate_element(name=tag)
    for language in obj:
        et_lss.append(_generate_element(name=NS_AAS + "langString",
                                        text=obj[language],
                                        attributes={"lang": language}))
    return et_lss


def administrative_information_to_xml(obj: model.AdministrativeInformation,
                                      tag: str = NS_AAS+"administration") -> etree.Element:
    """
    serialization of objects of class AdministrativeInformation to XML

    :param obj: object of class AdministrativeInformation
    :param tag: tag of the serialized element. default is "administration"
    :return: serialized ElementTree object
    """
    et_administration = _generate_element(tag)
    if obj.version:
        et_administration.append(_generate_element(name=NS_AAS + "version", text=obj.version))
        if obj.revision:
            et_administration.append(_generate_element(name=NS_AAS + "revision", text=obj.revision))
    return et_administration


def data_element_to_xml(obj: model.DataElement) -> etree.Element:
    """
    serialization of objects of class DataElement to XML

    :param obj: Object of class DataElement
    :return: serialized ElementTree element
    """
    if isinstance(obj, model.MultiLanguageProperty):
        return multi_language_property_to_xml(obj)
    if isinstance(obj, model.Property):
        return property_to_xml(obj)
    if isinstance(obj, model.Range):
        return range_to_xml(obj)
    if isinstance(obj, model.Blob):
        return blob_to_xml(obj)
    if isinstance(obj, model.File):
        return file_to_xml(obj)
    if isinstance(obj, model.ReferenceElement):
        return reference_element_to_xml(obj)


def reference_to_xml(obj: model.Reference, tag: str = NS_AAS+"reference") -> etree.Element:
    """
    serialization of objects of class Reference to XML

    :param obj: object of class Reference
    :param tag: tag of the returned element
    :return: serialized ElementTree
    """
    et_reference = _generate_element(tag)
    et_keys = _generate_element(name=NS_AAS + "keys")
    for aas_key in obj.key:
        et_keys.append(_generate_element(name=NS_AAS + "key",
                                         text=aas_key.value,
                                         attributes={"idType": _generic.KEY_TYPES[aas_key.id_type],
                                                     "local": boolean_to_xml(aas_key.local),
                                                     "type": _generic.KEY_ELEMENTS[aas_key.type]}))
    et_reference.append(et_keys)
    return et_reference


def formula_to_xml(obj: model.Formula, tag: str = NS_AAS+"formula") -> etree.Element:
    """
    serialization of objects of class Formula to XML

    :param obj: object of class Formula
    :param tag: tag of the ElementTree object, default is "formula"
    :return: serialized ElementTree object
    """
    et_formula = abstract_classes_to_xml(tag, obj)
    if obj.depends_on:
        et_depends_on = _generate_element(name=NS_AAS + "dependsOnRefs", text=None)
        for aas_reference in obj.depends_on:
            et_depends_on.append(reference_to_xml(aas_reference, NS_AAS+"reference"))
        et_formula.append(et_depends_on)
    return et_formula


def qualifier_to_xml(obj: model.Qualifier, tag: str = NS_AAS+"qualifier") -> etree.Element:
    """
    serialization of objects of class Qualifier to XML

    :param obj: object of class Qualifier
    :param tag: tag of the serialized ElementTree object, default is "qualifier"
    :return: serialized ElementTreeObject
    """
    et_qualifier = abstract_classes_to_xml(tag, obj)
    et_qualifier.append(_generate_element(NS_AAS + "type", text=obj.type))
    et_qualifier.append(_generate_element(NS_AAS + "valueType", text=model.datatypes.XSD_TYPE_NAMES[obj.value_type]))
    if obj.value_id:
        et_qualifier.append(reference_to_xml(obj.value_id, NS_AAS+"valueId"))
    if obj.value:
        et_qualifier.append(_value_to_xml(obj.value, obj.value_type))
    return et_qualifier


def value_reference_pair_to_xml(obj: model.ValueReferencePair,
                                tag: str = NS_AAS+"valueReferencePair") -> etree.Element:
    """
    serialization of objects of class ValueReferencePair to XML

    todo: couldn't find it in the official schema, so guessing how to implement serialization
          check namespace, tag and correct serialization

    :param obj: object of class ValueReferencePair
    :param tag: tag of the serialized element, default is "valueReferencePair"
    :return: serialized ElementTree object
    """
    et_vrp = _generate_element(tag)
    et_vrp.append(_value_to_xml(obj.value, obj.value_type))
    et_vrp.append(reference_to_xml(obj.value_id, "valueId"))
    return et_vrp


def value_list_to_xml(obj: model.ValueList,
                      tag: str = NS_AAS+"valueList") -> etree.Element:
    """
    serialization of objects of class ValueList to XML

    todo: couldn't find it in the official schema, so guessing how to implement serialization

    :param obj: object of class ValueList
    :param tag: tag of the serialized element, default is "valueList"
    :return: serialized ElementTree object
    """
    et_value_list = _generate_element(tag)
    for aas_reference_pair in obj:
        et_value_list.append(value_reference_pair_to_xml(aas_reference_pair, "valueReferencePair"))
    return et_value_list


# ##############################################################
# transformation functions to serialize classes from model.aas
# ##############################################################


def view_to_xml(obj: model.View, tag: str = NS_AAS+"view") -> etree.Element:
    """
    serialization of objects of class View to XML

    :param obj: object of class View
    :param tag: namespace+tag of the ElementTree object. default is "view"
    :return: serialized ElementTree object
    """
    et_view = abstract_classes_to_xml(tag, obj)
    et_contained_elements = _generate_element(name=NS_AAS + "containedElements")
    if obj.contained_element:
        for contained_element in obj.contained_element:
            et_contained_elements.append(reference_to_xml(contained_element, NS_AAS+"containedElementRef"))
    et_view.append(et_contained_elements)
    return et_view


def asset_to_xml(obj: model.Asset, tag: str = NS_AAS+"asset") -> etree.Element:
    """
    serialization of objects of class Asset to XML

    :param obj: object of class Asset
    :param tag: namespace+tag of the ElementTree object. default is "asset"
    :return: serialized ElementTree object
    """
    et_asset = abstract_classes_to_xml(tag, obj)
    if obj.asset_identification_model:
        et_asset.append(reference_to_xml(obj.asset_identification_model, NS_AAS+"assetIdentificationModelRef"))
    if obj.bill_of_material:
        et_asset.append(reference_to_xml(obj.bill_of_material, NS_AAS+"billOfMaterialRef"))
    et_asset.append(_generate_element(name=NS_AAS + "kind", text=_generic.ASSET_KIND[obj.kind]))
    return et_asset


def concept_description_to_xml(obj: model.ConceptDescription,
                               tag: str = NS_AAS+"conceptDescription") -> etree.Element:
    """
    serialization of objects of class ConceptDescription to XML

    :param obj: object of class ConceptDescription
    :param tag: tag of the ElementTree object. default is "conceptDescription"
    :return: serialized ElementTree object
    """
    et_concept_description = abstract_classes_to_xml(tag, obj)
    if isinstance(obj, model.concept.IEC61360ConceptDescription):
        et_embedded_data_specification = _generate_element(NS_AAS+"embeddedDataSpecification")
        et_data_spec_content = _generate_element(NS_AAS+"dataSpecificationContent")
        et_data_spec_content.append(_iec61360_concept_description_to_xml(obj))
        et_embedded_data_specification.append(et_data_spec_content)
        et_concept_description.append(et_embedded_data_specification)
        et_embedded_data_specification.append(reference_to_xml(model.Reference(tuple([model.Key(
            model.KeyElements.GLOBAL_REFERENCE,
            False,
            "http://admin-shell.io/DataSpecificationTemplates/DataSpecificationIEC61360/2/0",
            model.KeyType.IRI
        )])), NS_AAS+"dataSpecification"))
    if obj.is_case_of:
        for reference in obj.is_case_of:
            et_concept_description.append(reference_to_xml(reference, NS_AAS+"isCaseOf"))
    return et_concept_description


def _iec61360_concept_description_to_xml(obj: model.concept.IEC61360ConceptDescription,
                                         tag: str = NS_AAS+"dataSpecificationIEC61360") -> etree.Element:
    """
    Add the 'embeddedDataSpecifications' attribute to IEC61360ConceptDescription's JSON representation.

    `IEC61360ConceptDescription` is not a distinct class according DotAAS, but instead is built by referencing
    "DataSpecificationIEC61360" as dataSpecification. However, we implemented it as an explicit class, inheriting from
    ConceptDescription, but we want to generate compliant XML documents. So, we fake the XML structure of an object
    with dataSpecifications.

    :param obj: model.concept.IEC61360ConceptDescription object
    :param tag: name of the serialized lss_tag
    :return: serialized ElementTree object
    """

    def _iec_lang_string_set_to_xml(lss: model.LangStringSet, lss_tag: str) -> etree.Element:
        """
        serialization of objects of class LangStringSet to XML

        :param lss: object of class LangStringSet
        :param lss_tag: lss_tag name of the returned XML element (incl. namespace)
        :return: serialized ElementTree object
        """
        et_lss = _generate_element(name=lss_tag)
        for language in lss:
            et_lss.append(_generate_element(name=NS_IEC + "langString",
                                            text=lss[language],
                                            attributes={"lang": language}))
        return et_lss

    def _iec_reference_to_xml(ref: model.Reference, ref_tag: str = NS_AAS + "reference") -> etree.Element:
        """
        serialization of objects of class Reference to XML

        :param ref: object of class Reference
        :param ref_tag: ref_tag of the returned element
        :return: serialized ElementTree
        """
        et_reference = _generate_element(ref_tag)
        et_keys = _generate_element(name=NS_IEC + "keys")
        for aas_key in ref.key:
            et_keys.append(_generate_element(name=NS_IEC + "key",
                                             text=aas_key.value,
                                             attributes={"idType": _generic.KEY_TYPES[aas_key.id_type],
                                                         "local": boolean_to_xml(aas_key.local),
                                                         "type": _generic.KEY_ELEMENTS[aas_key.type]}))
        et_reference.append(et_keys)
        return et_reference

    def _iec_value_reference_pair_to_xml(vrp: model.ValueReferencePair,
                                         vrp_tag: str = NS_IEC + "valueReferencePair") -> etree.Element:
        """
        serialization of objects of class ValueReferencePair to XML

        :param vrp: object of class ValueReferencePair
        :param vrp_tag: vl_tag of the serialized element, default is "valueReferencePair"
        :return: serialized ElementTree object
        """
        et_vrp = _generate_element(vrp_tag)
        et_vrp.append(_iec_reference_to_xml(vrp.value_id, NS_IEC + "valueId"))
        et_vrp.append(_value_to_xml(vrp.value, vrp.value_type, tag=NS_IEC+"value"))
        return et_vrp

    def _iec_value_list_to_xml(vl: model.ValueList,
                               vl_tag: str = NS_IEC + "valueList") -> etree.Element:
        """
        serialization of objects of class ValueList to XML

        :param vl: object of class ValueList
        :param vl_tag: vl_tag of the serialized element, default is "valueList"
        :return: serialized ElementTree object
        """
        et_value_list = _generate_element(vl_tag)
        for aas_reference_pair in vl:
            et_value_list.append(_iec_value_reference_pair_to_xml(aas_reference_pair, NS_IEC+"valueReferencePair"))
        return et_value_list

    et_iec = _generate_element(tag)
    et_iec.append(_iec_lang_string_set_to_xml(obj.preferred_name, NS_IEC + "preferredName"))
    if obj.short_name:
        et_iec.append(_iec_lang_string_set_to_xml(obj.short_name, NS_IEC + "shortName"))
    if obj.unit:
        et_iec.append(_generate_element(NS_IEC+"unit", text=obj.unit))
    if obj.unit_id:
        et_iec.append(_iec_reference_to_xml(obj.unit_id, NS_IEC+"unitId"))
    if obj.source_of_definition:
        et_iec.append(_generate_element(NS_IEC+"sourceOfDefinition", text=obj.source_of_definition))
    if obj.symbol:
        et_iec.append(_generate_element(NS_IEC+"symbol", text=obj.symbol))
    if obj.data_type:
        et_iec.append(_generate_element(NS_IEC+"dataType", text=_generic.IEC61360_DATA_TYPES[obj.data_type]))
    if obj.definition:
        et_iec.append(_iec_lang_string_set_to_xml(obj.definition, NS_IEC + "definition"))
    if obj.value_format:
        et_iec.append(_generate_element(NS_IEC+"valueFormat", text=model.datatypes.XSD_TYPE_NAMES[obj.value_format]))
    if obj.value_list:
        et_iec.append(_iec_value_list_to_xml(obj.value_list, NS_IEC+"valueList"))
    if obj.value:
        et_iec.append(_generate_element(NS_IEC+"value", text=model.datatypes.xsd_repr(obj.value)))
    if obj.value_id:
        et_iec.append(_iec_reference_to_xml(obj.value_id, NS_IEC+"valueId"))
    if obj.level_types:
        for level_type in obj.level_types:
            et_iec.append(_generate_element(NS_IEC+"levelType", text=_generic.IEC61360_LEVEL_TYPES[level_type]))
    return et_iec


def concept_dictionary_to_xml(obj: model.ConceptDictionary,
                              tag: str = NS_AAS+"conceptDictionary") -> etree.Element:
    """
    serialization of objects of class ConceptDictionary to XML

    :param obj: object of class ConceptDictionary
    :param tag: tag of the ElementTree object. default is "conceptDictionary"
    :return: serialized ElementTree object
    """
    et_concept_dictionary = abstract_classes_to_xml(tag, obj)
    et_concept_descriptions_refs = _generate_element(NS_AAS + "conceptDescriptionRefs")
    if obj.concept_description:
        for reference in obj.concept_description:
            et_concept_descriptions_refs.append(reference_to_xml(reference, NS_AAS+"conceptDescriptionRef"))
    et_concept_dictionary.append(et_concept_descriptions_refs)
    return et_concept_dictionary


def asset_administration_shell_to_xml(obj: model.AssetAdministrationShell,
                                      tag: str = NS_AAS+"assetAdministrationShell") -> etree.Element:
    """
    serialization of objects of class AssetAdministrationShell to XML

    :param obj: object of class AssetAdministrationShell
    :param tag: tag of the ElementTree object. default is "assetAdministrationShell"
    :return: serialized ElementTree object
    """
    et_aas = abstract_classes_to_xml(tag, obj)
    if obj.derived_from:
        et_aas.append(reference_to_xml(obj.derived_from, tag=NS_AAS+"derivedFrom"))
    et_aas.append(reference_to_xml(obj.asset, tag=NS_AAS+"assetRef"))
    if obj.submodel:
        et_submodels = _generate_element(NS_AAS + "submodelRefs")
        for reference in obj.submodel:
            et_submodels.append(reference_to_xml(reference, tag=NS_AAS+"submodelRef"))
        et_aas.append(et_submodels)
    if obj.view:
        et_views = _generate_element(NS_AAS + "views")
        for view in obj.view:
            et_views.append(view_to_xml(view, NS_AAS+"view"))
        et_aas.append(et_views)
    if obj.concept_dictionary:
        et_concept_dictionaries = _generate_element(NS_AAS + "conceptDictionaries")
        for concept_dictionary in obj.concept_dictionary:
            et_concept_dictionaries.append(concept_dictionary_to_xml(concept_dictionary,
                                                                     NS_AAS+"conceptDictionary"))
        et_aas.append(et_concept_dictionaries)
    if obj.security:
        et_aas.append(security_to_xml(obj.security, tag=NS_ABAC+"security"))
    return et_aas


# ##############################################################
# transformation functions to serialize classes from model.security
# ##############################################################


def security_to_xml(obj: model.Security,
                    tag: str = NS_ABAC+"security") -> etree.Element:
    """
    serialization of objects of class Security to XML

    todo: This is not yet implemented

    :param obj: object of class Security
    :param tag: tag of the serialized element (optional). Default is "security"
    :return: serialized ElementTree object
    """
    return abstract_classes_to_xml(tag, obj)


# ##############################################################
# transformation functions to serialize classes from model.submodel
# ##############################################################


def submodel_element_to_xml(obj: model.SubmodelElement) -> etree.Element:
    """
    serialization of objects of class SubmodelElement to XML

    :param obj: object of class SubmodelElement
    :return: serialized ElementTree object
    """
    if isinstance(obj, model.DataElement):
        return data_element_to_xml(obj)
    if isinstance(obj, model.BasicEvent):
        return basic_event_to_xml(obj)
    if isinstance(obj, model.Capability):
        return capability_to_xml(obj)
    if isinstance(obj, model.Entity):
        return entity_to_xml(obj)
    if isinstance(obj, model.Operation):
        return operation_to_xml(obj)
    if isinstance(obj, model.AnnotatedRelationshipElement):
        return annotated_relationship_element_to_xml(obj)
    if isinstance(obj, model.RelationshipElement):
        return relationship_element_to_xml(obj)
    if isinstance(obj, model.SubmodelElementCollection):
        return submodel_element_collection_to_xml(obj)


def submodel_to_xml(obj: model.Submodel,
                    tag: str = NS_AAS+"submodel") -> etree.Element:
    """
    serialization of objects of class Submodel to XML

    :param obj: object of class Submodel
    :param tag: tag of the serialized element (optional). Default is "submodel"
    :return: serialized ElementTree object
    """
    et_submodel = abstract_classes_to_xml(tag, obj)
    et_submodel_elements = _generate_element(NS_AAS + "submodelElements")
    if obj.submodel_element:
        for submodel_element in obj.submodel_element:
            # TODO: simplify this should our suggestion regarding the XML schema get accepted
            # https://git.rwth-aachen.de/acplt/pyi40aas/-/issues/57
            et_submodel_element = _generate_element(NS_AAS+"submodelElement")
            et_submodel_element.append(submodel_element_to_xml(submodel_element))
            et_submodel_elements.append(et_submodel_element)
    et_submodel.append(et_submodel_elements)
    return et_submodel


def property_to_xml(obj: model.Property,
                    tag: str = NS_AAS+"property") -> etree.Element:
    """
    serialization of objects of class Property to XML

    :param obj: object of class Property
    :param tag: tag of the serialized element (optional), default is "property"
    :return: serialized ElementTree object
    """
    et_property = abstract_classes_to_xml(tag, obj)
    et_property.append(_generate_element(NS_AAS + "valueType", text=model.datatypes.XSD_TYPE_NAMES[obj.value_type]))
    if obj.value:
        et_property.append(_value_to_xml(obj.value, obj.value_type))
    if obj.value_id:
        et_property.append(reference_to_xml(obj.value_id, NS_AAS + "valueId"))
    return et_property


def multi_language_property_to_xml(obj: model.MultiLanguageProperty,
                                   tag: str = NS_AAS+"multiLanguageProperty") -> etree.Element:
    """
    serialization of objects of class MultiLanguageProperty to XML

    :param obj: object of class MultiLanguageProperty
    :param tag: tag of the serialized element (optional), default is "multiLanguageProperty"
    :return: serialized ElementTree object
    """
    et_multi_language_property = abstract_classes_to_xml(tag, obj)
    if obj.value_id:
        et_multi_language_property.append(reference_to_xml(obj.value_id, NS_AAS+"valueId"))
    if obj.value:
        et_multi_language_property.append(lang_string_set_to_xml(obj.value, tag=NS_AAS + "value"))
    return et_multi_language_property


def range_to_xml(obj: model.Range,
                 tag: str = NS_AAS+"range") -> etree.Element:
    """
    serialization of objects of class Range to XML

    :param obj: object of class Range
    :param tag: namespace+tag of the serialized element (optional), default is "range
    :return: serialized ElementTree object
    """
    et_range = abstract_classes_to_xml(tag, obj)
    et_range.append(_generate_element(name=NS_AAS + "valueType",
                                      text=model.datatypes.XSD_TYPE_NAMES[obj.value_type]))
    if obj.min is not None:
        et_range.append(_value_to_xml(obj.min, obj.value_type, tag=NS_AAS+"min"))
    if obj.max is not None:
        et_range.append(_value_to_xml(obj.max, obj.value_type, tag=NS_AAS+"max"))
    return et_range


def blob_to_xml(obj: model.Blob,
                tag: str = NS_AAS+"blob") -> etree.Element:
    """
    serialization of objects of class Blob to XML

    :param obj: object of class Blob
    :param tag: tag of the serialized element, default is "blob"
    :return: serialized ElementTree object
    """
    et_blob = abstract_classes_to_xml(tag, obj)
    et_value = etree.Element(NS_AAS + "value")
    if obj.value is not None:
        et_value.text = base64.b64encode(obj.value).decode()
    et_blob.append(et_value)
    et_blob.append(_generate_element(NS_AAS + "mimeType", text=obj.mime_type))
    return et_blob


def file_to_xml(obj: model.File,
                tag: str = NS_AAS+"file") -> etree.Element:
    """
    serialization of objects of class File to XML

    :param obj: object of class File
    :param tag: tag of the serialized element, default is "file"
    :return: serialized ElementTree object
    """
    et_file = abstract_classes_to_xml(tag, obj)
    et_file.append(_generate_element(NS_AAS + "mimeType", text=obj.mime_type))
    if obj.value:
        et_file.append(_generate_element(NS_AAS + "value", text=obj.value))
    return et_file


def reference_element_to_xml(obj: model.ReferenceElement,
                             tag: str = NS_AAS+"referenceElement") -> etree.Element:
    """
    serialization of objects of class ReferenceElement to XMl

    :param obj: object of class ReferenceElement
    :param tag: namespace+tag of the serialized element (optional), default is "referenceElement"
    :return: serialized ElementTree object
    """
    et_reference_element = abstract_classes_to_xml(tag, obj)
    if obj.value:
        et_reference_element.append(reference_to_xml(obj.value, NS_AAS+"value"))
    return et_reference_element


def submodel_element_collection_to_xml(obj: model.SubmodelElementCollection,
                                       tag: str = NS_AAS+"submodelElementCollection") -> etree.Element:
    """
    serialization of objects of class SubmodelElementCollection to XML

    Note that we do not have parameter "allowDuplicates" in out implementation

    :param obj: object of class SubmodelElementCollection
    :param tag: namespace+tag of the serialized element (optional), default is "submodelElementCollection"
    :return: serialized ElementTree object
    """
    et_submodel_element_collection = abstract_classes_to_xml(tag, obj)
    # todo: remove wrapping submodelElement-tag, in accordance to future schema
    et_value = _generate_element(NS_AAS + "value")
    if obj.value:
        for submodel_element in obj.value:
            et_submodel_element = _generate_element(NS_AAS+"submodelElement")
            et_submodel_element.append(submodel_element_to_xml(submodel_element))
            et_value.append(et_submodel_element)
    et_submodel_element_collection.append(et_value)
    et_submodel_element_collection.append(_generate_element(NS_AAS + "ordered", text=boolean_to_xml(obj.ordered)))
    et_submodel_element_collection.append(_generate_element(NS_AAS + "allowDuplicates", text="false"))
    return et_submodel_element_collection


def relationship_element_to_xml(obj: model.RelationshipElement,
                                tag: str = NS_AAS+"relationshipElement") -> etree.Element:
    """
    serialization of objects of class RelationshipElement to XML

    :param obj: object of class RelationshipElement
    :param tag: tag of the serialized element (optional), default is "relationshipElement"
    :return: serialized ELementTree object
    """
    et_relationship_element = abstract_classes_to_xml(tag, obj)
    et_relationship_element.append(reference_to_xml(obj.first, NS_AAS+"first"))
    et_relationship_element.append(reference_to_xml(obj.second, NS_AAS+"second"))
    return et_relationship_element


def annotated_relationship_element_to_xml(obj: model.AnnotatedRelationshipElement,
                                          tag: str = NS_AAS+"annotatedRelationshipElement") -> etree.Element:
    """
    serialization of objects of class AnnotatedRelationshipElement to XML

    :param obj: object of class AnnotatedRelationshipElement
    :param tag: tag of the serialized element (optional), default is "annotatedRelationshipElement
    :return: serialized ElementTree object
    """
    et_annotated_relationship_element = relationship_element_to_xml(obj, tag)
    et_annotations = _generate_element(name=NS_AAS+"annotations")
    if obj.annotation:
        for data_element in obj.annotation:
            et_data_element = _generate_element(name=NS_AAS+"dataElement")
            et_data_element.append(data_element_to_xml(data_element))
            et_annotations.append(et_data_element)
    et_annotated_relationship_element.append(et_annotations)
    return et_annotated_relationship_element


def operation_variable_to_xml(obj: model.OperationVariable,
                              tag: str = NS_AAS+"operationVariable") -> etree.Element:
    """
    serialization of objects of class OperationVariable to XML

    :param obj: object of class OperationVariable
    :param tag: tag of the serialized element (optional), default is "operationVariable"
    :return: serialized ElementTree object
    """
    et_operation_variable = _generate_element(tag)
    et_value = _generate_element(NS_AAS+"value")
    et_value.append(submodel_element_to_xml(obj.value))
    et_operation_variable.append(et_value)
    return et_operation_variable


def operation_to_xml(obj: model.Operation,
                     tag: str = NS_AAS+"operation") -> etree.Element:
    """
    serialization of objects of class Operation to XML

    :param obj: object of class Operation
    :param tag: namespace+tag of the serialized element (optional), default is "operation"
    :return: serialized ElementTree object
    """
    et_operation = abstract_classes_to_xml(tag, obj)
    if obj.input_variable:
        for input_ov in obj.input_variable:
            et_operation.append(operation_variable_to_xml(input_ov, NS_AAS+"inputVariable"))
    if obj.output_variable:
        for output_ov in obj.output_variable:
            et_operation.append(operation_variable_to_xml(output_ov, NS_AAS+"outputVariable"))
    if obj.in_output_variable:
        for in_out_ov in obj.in_output_variable:
            et_operation.append(operation_variable_to_xml(in_out_ov, NS_AAS+"inoutputVariable"))
    return et_operation


def capability_to_xml(obj: model.Capability,
                      tag: str = NS_AAS+"capability") -> etree.Element:
    """
    serialization of objects of class Capability to XML

    :param obj: object of class Capability
    :param tag: tag of the serialized element, default is "capability"
    :return: serialized ElementTree object
    """
    return abstract_classes_to_xml(tag, obj)


def entity_to_xml(obj: model.Entity,
                  tag: str = NS_AAS+"entity") -> etree.Element:
    """
    serialization of objects of class Entity to XML

    :param obj: object of class Entity
    :param tag: tag of the serialized element (optional), default is "entity"
    :return: serialized ElementTree object
    """
    # todo: remove wrapping submodelElement, in accordance to future schemas
    et_entity = abstract_classes_to_xml(tag, obj)
    et_statements = _generate_element(NS_AAS + "statements")
    for statement in obj.statement:
        # todo: remove the <submodelElement> once the proposed changes get accepted
        et_submodel_element = _generate_element(NS_AAS+"submodelElement")
        et_submodel_element.append(submodel_element_to_xml(statement))
        et_statements.append(et_submodel_element)
    et_entity.append(et_statements)
    et_entity.append(_generate_element(NS_AAS + "entityType", text=_generic.ENTITY_TYPES[obj.entity_type]))
    if obj.asset:
        et_entity.append(reference_to_xml(obj.asset, NS_AAS+"assetRef"))
    return et_entity


def basic_event_to_xml(obj: model.BasicEvent,
                       tag: str = NS_AAS+"basicEvent") -> etree.Element:
    """
    serialization of objects of class BasicEvent to XML

    :param obj: object of class BasicEvent
    :param tag: tag of the serialized element (optional), default is "basicEvent"
    :return: serialized ElementTree object
    """
    et_basic_event = abstract_classes_to_xml(tag, obj)
    et_basic_event.append(reference_to_xml(obj.observed, NS_AAS+"observed"))
    return et_basic_event


# ##############################################################
# general functions
# ##############################################################


def write_aas_xml_file(file: IO,
                       data: model.AbstractObjectStore,
                       **kwargs) -> None:
    """
    Write a set of AAS objects to an Asset Administration Shell XML file according to 'Details of the Asset
    Administration Shell', chapter 5.4

    :param file: A file-like object to write the XML-serialized data to
    :param data: ObjectStore which contains different objects of the AAS meta model which should be serialized to an
                 XML file
    :param kwargs: Additional keyword arguments to be passed to tree.write()
    """
    # separate different kind of objects
    assets = []
    asset_administration_shells = []
    submodels = []
    concept_descriptions = []
    for obj in data:
        if isinstance(obj, model.Asset):
            assets.append(obj)
        if isinstance(obj, model.AssetAdministrationShell):
            asset_administration_shells.append(obj)
        if isinstance(obj, model.Submodel):
            submodels.append(obj)
        if isinstance(obj, model.ConceptDescription):
            concept_descriptions.append(obj)

    # serialize objects to XML
    root = etree.Element(NS_AAS + "aasenv", nsmap=NS_MAP)
    et_asset_administration_shells = etree.Element(NS_AAS + "assetAdministrationShells")
    for aas_obj in asset_administration_shells:
        et_asset_administration_shells.append(asset_administration_shell_to_xml(aas_obj))
    et_assets = _generate_element(NS_AAS + "assets")
    for ass_obj in assets:
        et_assets.append(asset_to_xml(ass_obj))
    et_submodels = etree.Element(NS_AAS + "submodels")
    for sub_obj in submodels:
        et_submodels.append(submodel_to_xml(sub_obj))
    et_concept_descriptions = etree.Element(NS_AAS + "conceptDescriptions")
    for con_obj in concept_descriptions:
        et_concept_descriptions.append(concept_description_to_xml(con_obj))
    root.insert(0, et_concept_descriptions)
    root.insert(0, et_submodels)
    root.insert(0, et_assets)
    root.insert(0, et_asset_administration_shells)

    tree = etree.ElementTree(root)
    tree.write(file, encoding="UTF-8", xml_declaration=True, method="xml", **kwargs)



# Copyright (c) 2020 PyI40AAS Contributors
#
# This program and the accompanying materials are made available under the terms of the Eclipse Public License v. 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0 which is available
# at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
"""
This module implements the basic structures of the AAS metamodel, including the abstract classes and enums needed for
the higher level classes to inherit from.
"""

import abc
import inspect
import itertools
from enum import Enum, unique
from typing import List, Optional, Set, TypeVar, MutableSet, Generic, Iterable, Dict, Iterator, Union, overload, \
    MutableSequence, Type, Any, TYPE_CHECKING, Tuple
import re

from . import datatypes
from ..backend import backends

if TYPE_CHECKING:
    from . import provider

DataTypeDef = Type[datatypes.AnyXSDType]
ValueDataType = datatypes.AnyXSDType  # any xsd atomic type (from .datatypes)
BlobType = bytes
MimeType = str  # any mimetype as in RFC2046
PathType = str
QualifierType = str
# A dict of language-Identifier (according to ISO 639-1 and ISO 3166-1) and string in this language.
# The meaning of the string in each language is the same.
# << Data Type >> Example ["en-US", "germany"]
LangStringSet = Dict[str, str]


@unique
class IdentifierType(Enum):
    """
    Enumeration of different types of Identifiers for global identification

    :cvar IRDI: IRDI (International Registration Data Identifier) according to ISO29002-5 as an Identifier scheme for
                properties and classifications.
    :cvar IRI: IRI according to Rfc 3987. Every URI is an IRI
    :cvar CUSTOM: Custom identifiers like GUIDs (globally unique Identifiers)
    """

    IRDI = 0
    IRI = 1
    CUSTOM = 2


@unique
class KeyElements(Enum):
    """
    Enumeration for denoting which kind of entity is referenced. They can be categorized in ReferableElements,
    IdentifiableElements and other KeyElements

    # IdentifiableElements starting from 0
    :cvar ASSET: asset
    :cvar ASSET_ADMINISTRATION_SHELL: asset administration shell
    :cvar CONCEPT_DESCRIPTION: concept description
    :cvar SUBMODEL: submodel

    # ReferableElements starting from 1000
    :cvar ACCESS_PERMISSION_RULE: access permission rule
    :cvar ANNOTATED_RELATIONSHIP_ELEMENT: annotated relationship element
    :cvar BASIC_EVENT: basic event
    :cvar BLOB: blob
    :cvar CAPABILITY: capability
    :cvar CONCEPT_DICTIONARY: concept dictionary
    :cvar DATA_ELEMENT: data element,
                        Note: Date Element is abstract, i. e. if a key uses "DATA_ELEMENT" the reference may be
                              Property, File etc.
    :cvar ENTITY: entity
    :cvar EVENT: event, Note: Event is abstract
    :cvar FILE: file
    :cvar MULTI_LANGUAGE_PROPERTY: property with a value that can be provided in multiple languages
    :cvar OPERATION: operation
    :cvar PROPERTY: property
    :cvar RANGE: range with min and max
    :cvar REFERENCE_ELEMENT: reference
    :cvar RELATIONSHIP_ELEMENT: relationship
    :cvar SUBMODEL_ELEMENT: submodel element,
                            Note: Submodel Element is abstract, i.e. if a key uses “SUBMODEL_ELEMENT” the reference may
                                  be a Property, a SubmodelElementCollection, an Operation etc.
    :cvar SUBMODEL_ELEMENT_COLLECTION: collection of submodel elements
    :cvar VIEW: view

    # KeyElements starting from 2000
    :cvar GLOBAL_REFERENCE: reference to an element not belonging to an asset administration shel
    :cvar FRAGMENT_REFERENCE: unique reference to an element within a file. The file itself is assumed to be part of an
                             asset administration shell.
    """

    # IdentifiableElements starting from 0
    ASSET = 0
    ASSET_ADMINISTRATION_SHELL = 1
    CONCEPT_DESCRIPTION = 2
    SUBMODEL = 3

    # ReferableElements starting from 1000
    ACCESS_PERMISSION_RULE = 1000
    ANNOTATED_RELATIONSHIP_ELEMENT = 1001
    BASIC_EVENT = 1002
    BLOB = 1003
    CAPABILITY = 1004
    CONCEPT_DICTIONARY = 1005
    DATA_ELEMENT = 1006
    ENTITY = 1007
    EVENT = 1008
    FILE = 1009
    MULTI_LANGUAGE_PROPERTY = 1010
    OPERATION = 1011
    PROPERTY = 1012
    RANGE = 1013
    REFERENCE_ELEMENT = 1014
    RELATIONSHIP_ELEMENT = 1015
    SUBMODEL_ELEMENT = 1016
    SUBMODEL_ELEMENT_COLLECTION = 1017
    VIEW = 1018

    # KeyElements starting from 2000
    GLOBAL_REFERENCE = 2000
    FRAGMENT_REFERENCE = 2001


@unique
class KeyType(Enum):
    """
    Enumeration for denoting the type of the key value.

    :cvar IRDI: IRDI (International Registration Data Identifier) according to ISO29002-5 as an Identifier scheme for
                properties and classifications.
    :cvar IRI: IRI according to Rfc 3987. Every URI is an IRI
    :cvar CUSTOM: Custom identifiers like GUIDs (globally unique Identifiers)
    :cvar IDSHORT: id_short of a referable element
    :cvar FRAGMENT_ID: identifier of a fragment within a file
    """

    IRDI = 0
    IRI = 1
    CUSTOM = 2
    IDSHORT = 3
    FRAGMENT_ID = 4

    @property
    def is_local_key_type(self) -> bool:
        return self in (KeyType.IDSHORT, KeyType.FRAGMENT_ID)


@unique
class EntityType(Enum):
    """
    Enumeration for denoting whether an entity is a self-managed or a co-managed entity

    :cvar CO_MANAGED_ENTITY: For co-managed entities there is no separate AAS. Co-managed entities need to be part of a
                             self-managed entity
    :cvar SELF_MANAGED_ENTITY: Self-managed entities have their own AAS but can be part of the bill of material of a
                               composite self-managed entity. The asset of an I4.0-component is a self-managed entity
                               per definition.
    """

    CO_MANAGED_ENTITY = 0
    SELF_MANAGED_ENTITY = 1


@unique
class ModelingKind(Enum):
    """
    Enumeration for denoting whether an element is a type or an instance.

    :cvar TEMPLATE: Software element which specifies the common attributes shared by all instances of the template
    :cvar INSTANCE: concrete, clearly identifiable component of a certain template,
                    Note: It becomes an individual entity of a template, for example a device model, by defining
                          specific property values.
                    Note: In an object oriented view, an instance denotes an object of a template (class).
    """

    TEMPLATE = 0
    INSTANCE = 1


@unique
class AssetKind(Enum):
    """
    Enumeration for denoting whether an element is a type or an instance.

    :cvar TYPE: hardware or software element which specifies the common attributes shared by all instances of the type
    :cvar INSTANCE: concrete, clearly identifiable component of a certain type,
                    Note: It becomes an individual entity of a type, for example a device, by defining specific
                          property values.
                    Note: In an object oriented view, an instance denotes an object of a class (of a type)
    """

    TYPE = 0
    INSTANCE = 1


class Key:
    """
    A key is a reference to an element by its id.

    :ivar type: Denote which kind of entity is referenced. In case type = GlobalReference then the element is a
                global unique id. In all other cases the key references a model element of the same or of another AAS.
                The name of the model element is explicitly listed.
    :ivar local: Denotes if the key references a model element of the same AAS (=true) or not (=false). In case of
                 local = false the key may reference a model element of another AAS or an entity outside any AAS that
                 has a global unique id.
    :ivar value: The key value, for example an IRDI if the idType=IRDI
    :ivar id_type: Type of the key value. In case of idType = idShort local shall be true. In case type=GlobalReference
                   idType shall not be IdShort.
    """

    def __init__(self,
                 type_: KeyElements,
                 local: bool,
                 value: str,
                 id_type: KeyType):
        """
        Initializer of Key

        :param type_: Denote which kind of entity is referenced. In case type = GlobalReference then the element is a
                      global unique id. In all other cases the key references a model element of the same or of another
                      AAS. The name of the model element is explicitly listed.
        :param local: Denotes if the key references a model element of the same AAS (=true) or not (=false). In case of
                      local = false the key may reference a model element of another AAS or an entity outside any AAS
                      that has a global unique id.
        :param value: The key value, for example an IRDI if the idType=IRDI
        :param id_type: Type of the key value. In case of idType = idShort local shall be true. In case
                        type=GlobalReference idType shall not be IdShort.

        TODO: Add instruction what to do after construction
        """
        self.type: KeyElements
        self.local: bool
        self.value: str
        self.id_type: KeyType
        super().__setattr__('type', type_)
        super().__setattr__('local', local)
        super().__setattr__('value', value)
        super().__setattr__('id_type', id_type)

    def __setattr__(self, key, value):
        """Prevent modification of attributes."""
        raise AttributeError('Reference is immutable')

    def __repr__(self) -> str:
        return "Key(local={}, id_type={}, value={})".format(self.local, self.id_type.name, self.value)

    def __str__(self) -> str:
        return "{}={}".format(self.id_type.name, self.value)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Key):
            return NotImplemented
        return (self.id_type is other.id_type
                and self.value == other.value
                and self.local == other.local
                and self.type == other.type)

    def __hash__(self):
        return hash((self.id_type, self.value, self.local, self.type))

    def get_identifier(self) -> Optional["Identifier"]:
        """
        Get an identifier object corresponding to this key, if it is a global key.

        :return: None if this is no global key, otherwise a corresponding identifier object
        """
        if self.id_type.is_local_key_type:
            return None
        return Identifier(self.value, IdentifierType(self.id_type.value))

    @staticmethod
    def from_referable(referable: "Referable") -> "Key":
        """
        Construct a key for a given Referable (or Identifiable) object
        """
        # Get the `type` by finding the first class from the base classes list (via inspect.getmro), that is contained
        # in KEY_ELEMENTS_CLASSES
        from . import KEY_ELEMENTS_CLASSES
        try:
            key_type = next(iter(KEY_ELEMENTS_CLASSES[t]
                                 for t in inspect.getmro(type(referable))
                                 if t in KEY_ELEMENTS_CLASSES))
        except StopIteration:
            key_type = KeyElements.PROPERTY

        local = True  # TODO
        if isinstance(referable, Identifiable):
            return Key(key_type, local, referable.identification.id,
                       KeyType(referable.identification.id_type.value))
        else:
            return Key(key_type, local, referable.id_short, KeyType.IDSHORT)


class AdministrativeInformation:
    """
    Administrative meta-information for an element like version information.

    :ivar version: Version of the element.
    :ivar revision: Revision of the element.
    Constraint AASd-005: A revision requires a version. This means, if there is no version there is no revision
                         neither.
    """

    def __init__(self,
                 version: Optional[str] = None,
                 revision: Optional[str] = None):
        """
        Initializer of AdministrativeInformation

        :param version: Version of the element.
        :param revision: Revision of the element.

        Constraint AASd-005: A revision requires a version. This means, if there is no version there is no revision
                             neither.
        :raises ValueError: If version is None and revision is not None

        TODO: Add instruction what to do after construction
        """
        if version is None and revision is not None:
            raise ValueError("A revision requires a version. This means, if there is no version there is no revision "
                             "neither.")
        self.version: Optional[str] = version
        self._revision: Optional[str] = revision

    def _get_revision(self):
        return self._revision

    def _set_revision(self, revision: str):
        if self.version is None:
            raise ValueError("A revision requires a version. This means, if there is no version there is no revision "
                             "neither. Please set version first.")
        else:
            self._revision = revision

    def __eq__(self, other) -> bool:
        return self.version == other.version and self._revision == other._revision

    def __repr__(self) -> str:
        return "AdministrativeInformation(version={}, revision={})".format(self.version, self.revision)

    revision = property(_get_revision, _set_revision)


class Identifier:
    """
    Used to uniquely identify an entity by using an identifier.

    :ivar id: Identifier of the element. Its type is defined in id_type.
    :ivar id_type: Type of the Identifier, e.g. URI, IRDI etc. The supported Identifier types are defined in
                   the enumeration "IdentifierType".
    """

    def __init__(self,
                 id_: str,
                 id_type: IdentifierType):
        """
        Initializer of Identifier

        :param id_: Identifier of the element. Its type is defined in id_type.
        :param id_type: Type of the Identifier, e.g. URI, IRDI etc. The supported Identifier types are defined in the
                        enumeration "IdentifierType".

        TODO: Add instruction what to do after construction
        """
        self.id: str
        self.id_type: IdentifierType
        super().__setattr__('id', id_)
        super().__setattr__('id_type', id_type)

    def __setattr__(self, key, value):
        """Prevent modification of attributes."""
        raise AttributeError('Identifier are immutable')

    def __hash__(self):
        return hash((self.id_type, self.id))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Identifier):
            return NotImplemented
        return self.id_type == other.id_type and self.id == other.id

    def __repr__(self) -> str:
        return "Identifier({}={})".format(self.id_type.name, self.id)


class Referable(metaclass=abc.ABCMeta):
    """
    An element that is referable by its id_short. This id is not globally unique. This id is unique within
    the name space of the element.

    << abstract >>

    :ivar id_short: Identifying string of the element within its name space.
                    Constraint AASd-001: In case of a referable element not being an identifiable element this id is
                                         mandatory and used for referring to the element in its name space.
                    Constraint AASd-002: idShort shall only feature letters, digits, underscore ("_"); starting
                                         mandatory with a letter.
                    Constraint AASd-003: idShort shall be matched case insensitive.
    :ivar category: The category is a value that gives further meta information w.r.t. to the class of the element.
                    It affects the expected existence of attributes and the applicability of constraints.
    :ivar description: Description or comments on the element.
    :ivar parent: Reference to the next referable parent element of the element.
                  Constraint AASd-004: Add parent in case of non identifiable elements.
    :ivar source: Source of the object, an URI, that defines where this object's data originates from.
                  This is used to specify where the Referable should be updated from and committed to.
                  Default is an empty string, making it use the source of its ancestor, if possible.
    """
    @abc.abstractmethod
    def __init__(self):
        super().__init__()
        self._id_short: Optional[str] = ""
        self.category: Optional[str] = ""
        self.description: Optional[LangStringSet] = set()
        # We use a Python reference to the parent Namespace instead of a Reference Object, as specified. This allows
        # simpler and faster navigation/checks and it has no effect in the serialized data formats anyway.
        self.parent: Optional[Namespace] = None
        self.source: str = ""

    def __repr__(self) -> str:
        reversed_path = []
        item = self  # type: Any
        while item is not None:
            if isinstance(item, Identifiable):
                reversed_path.append(str(item.identification))
                break
            elif isinstance(item, Referable):
                reversed_path.append(item.id_short)
                item = item.parent
            else:
                raise AttributeError('Referable must have an identifiable as root object and only parents that are '
                                     'referable')

        return "{}[{}]".format(self.__class__.__name__, " / ".join(reversed(reversed_path)))

    def _get_id_short(self):
        return self._id_short

    def _set_id_short(self, id_short: Optional[str]):
        """
        Check the input string

        Constraint AASd-001: In case of a referable element not being an identifiable element this id is mandatory and
        used for referring to the element in its name space.
        Constraint AASd-002: idShort shall only feature letters, digits, underscore ('_'); starting mandatory with a
        letter

        Additionally check that the idShort is not already present in the same parent Namespace (if this object is
        already contained in a parent Namespace).

        :param id_short: Identifying string of the element within its name space
        :raises ValueError: if the constraint is not fulfilled
        :raises KeyError: if the new idShort causes a name collision in the parent Namespace
        """

        if id_short is None and not hasattr(self, 'identification'):
            raise ValueError("The id_short for not identifiable elements is mandatory")
        test_id_short: str = str(id_short)
        if not re.match("^[a-zA-Z0-9_]*$", test_id_short):
            raise ValueError("The id_short must contain only letters, digits and underscore")
        if not re.match("^([a-zA-Z].*|)$", test_id_short):
            raise ValueError("The id_short must start with a letter")

        if self.parent is not None and id_short != self.id_short:
            for set_ in self.parent.namespace_element_sets:
                if id_short in set_:
                    raise KeyError("Referable with id_short '{}' is already present in the parent Namespace"
                                   .format(id_short))
            for set_ in self.parent.namespace_element_sets:
                if self in set_:
                    set_.discard(self)
                    self._id_short = id_short
                    set_.add(self)
                    break
        # Redundant to the line above. However this way, we make sure that we really update the _id_short
        self._id_short = id_short

    def update(self,
               max_age: float = 0,
               recursive: bool = True,
               _indirect_source: bool = True) -> None:
        """
        Update the local Referable object from any underlying external data source, using an appropriate backend

        If there is no source given, it will find its next ancestor with a source and update from this source.
        If there is no source in any ancestor, this function will do nothing

        :param max_age: Maximum age of the local data in seconds. This method may return early, if the previous update
            of the object has been performed less than `max_age` seconds ago.
        :param recursive: Also call update on all children of this object. Default is True
        :param _indirect_source: Internal parameter to avoid duplicate updating.
        :raises backends.BackendError: If no appropriate backend or the data source is not available
        """
        # TODO consider max_age
        if not _indirect_source:
            # Update was already called on an ancestor of this Referable. Only update it, if it has its own source
            if self.source != "":
                backends.get_backend(self.source).update_object(updated_object=self,
                                                                store_object=self,
                                                                relative_path=[])

        else:
            # Try to find a valid source for this Referable
            if self.source != "":
                backends.get_backend(self.source).update_object(updated_object=self,
                                                                store_object=self,
                                                                relative_path=[])
            else:
                store_object, relative_path = self.find_source()
                if store_object and relative_path is not None:
                    backends.get_backend(store_object.source).update_object(updated_object=self,
                                                                            store_object=store_object,
                                                                            relative_path=list(relative_path))

        if recursive:
            # update all the children who have their own source
            if isinstance(self, Namespace):
                for namespace_set in self.namespace_element_sets:
                    for referable in namespace_set:
                        referable.update(max_age, recursive=True, _indirect_source=False)

    def find_source(self) -> Tuple[Optional["Referable"], Optional[List[str]]]:  # type: ignore
        """
        Finds the closest source in this objects ancestors. If there is no source, returns None

        :return: (The closest ancestor with a defined source, the relative path of id_shorts to that ancestor)
        """
        referable: Referable = self
        relative_path: List[str] = [self.id_short]
        while referable is not None:
            if referable.source != "":
                relative_path.reverse()
                return referable, relative_path
            if referable.parent:
                assert(isinstance(referable.parent, Referable))
                referable = referable.parent
                relative_path.append(referable.id_short)
                continue
            break
        return None, None

    def update_from(self, other: "Referable", update_source: bool = False):
        """
        Internal function to updates the object's attributes from another object of a similar type.

        This function should not be used directly. It is typically used by backend implementations (database adapters,
        protocol clients, etc.) to update the object's data, after `update()` has been called.

        :param other: The object to update from
        :param update_source: Update the source attribute with the other's source attribute. This is not propagated
                              recursively
        """
        for name, var in vars(other).items():
            # do not update the parent, namespace_element_sets or source (depending on update_source parameter)
            if name in ("parent", "namespace_element_sets") or name == "source" and not update_source:
                continue
            if isinstance(var, NamespaceSet):
                # update the elements of the NameSpaceSet
                vars(self)[name].update_nss_from(var)
            else:
                vars(self)[name] = var  # that variable is not a NameSpaceSet, so it isn't Referable

    def commit(self) -> None:
        """
        Transfer local changes on this object to all underlying external data sources.

        This function commits the current state of this object to its own and each external data source of its
        ancestors. If there is no source, this function will do nothing.
        """
        current_ancestor = self.parent
        relative_path: List[str] = [self.id_short]
        # Commit to all ancestors with sources
        while current_ancestor:
            assert(isinstance(current_ancestor, Referable))
            if current_ancestor.source != "":
                backends.get_backend(current_ancestor.source).commit_object(committed_object=self,
                                                                            store_object=current_ancestor,
                                                                            relative_path=list(relative_path))
            relative_path.insert(0, current_ancestor.id_short)
            current_ancestor = current_ancestor.parent
        # Commit to own source and check if there are children with sources to commit to
        self._direct_source_commit()

    def _direct_source_commit(self):
        """
        Commits children of an ancestor recursively, if they have a specific source given
        """
        if self.source != "":
            backends.get_backend(self.source).commit_object(committed_object=self,
                                                            store_object=self,
                                                            relative_path=[])

        if isinstance(self, Namespace):
            for namespace_set in self.namespace_element_sets:
                for referable in namespace_set:
                    referable._direct_source_commit()

    id_short = property(_get_id_short, _set_id_short)


_RT = TypeVar('_RT', bound=Referable)


class UnexpectedTypeError(TypeError):
    """
    Exception to be raised by Reference.resolve() if the retrieved object has not the expected type.

    :ivar value: The object of unexpected type
    """
    def __init__(self, value: Referable, *args):
        super().__init__(*args)
        self.value = value


class Reference:
    """
    Reference to either a model element of the same or another AAs or to an external entity.

    A reference is an ordered list of keys, each key referencing an element. The complete list of keys may for
    example be concatenated to a path that then gives unique access to an element or entity

    :ivar: key: Ordered list of unique reference in its name space, each key referencing an element. The complete
                list of keys may for example be concatenated to a path that then gives unique access to an element
                or entity.
    :ivar: type: The type of the referenced object (additional attribute, not from the AAS Metamodel)
    """

    def __init__(self,
                 key: Tuple[Key, ...]):
        """
        Initializer of Reference

        :param key: Ordered list of unique reference in its name space, each key referencing an element. The complete
                    list of keys may for example be concatenated to a path that then gives unique access to an element
                    or entity.

        TODO: Add instruction what to do after construction
        """
        self.key: Tuple[Key, ...]
        super().__setattr__('key', key)

    def __setattr__(self, key, value):
        """Prevent modification of attributes."""
        raise AttributeError('Reference is immutable')

    def __repr__(self) -> str:
        return "Reference(key={})".format(self.key)

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Reference):
            return NotImplemented
        if len(self.key) != len(other.key):
            return False
        return all(k1 == k2 for k1, k2 in zip(self.key, other.key))


class AASReference(Reference, Generic[_RT]):
    """
    Typed Reference to any referable Asset Administration Shell object.

    This is a special construct of the implementation to allow typed references and dereferencing.
    """
    def __init__(self,
                 key: Tuple[Key, ...],
                 target_type: Type[_RT]):
        """
        Initializer of AASReference

        :param key: Ordered list of unique reference in its name space, each key referencing an element. The complete
                    list of keys may for example be concatenated to a path that then gives unique access to an element
                    or entity.
        :param: type_: The type of the referenced object (additional parameter, not from the AAS Metamodel)

        TODO: Add instruction what to do after construction
        """
        # TODO check keys for validity. GlobalReference and Fragment-Type keys are not allowed here
        super().__init__(key)
        self.type: Type[_RT]
        object.__setattr__(self, 'type', target_type)

    def resolve(self, provider_: "provider.AbstractObjectProvider") -> _RT:
        """
        Follow the reference and retrieve the Referable object it points to

        :return: The referenced object (or a proxy object for it)
        :raises IndexError: If the list of keys is empty
        :raises TypeError: If one of the intermediate objects on the path is not a Namespace
        :raises UnexpectedTypeError: If the retrieved object is not of the expected type (or one of its subclasses). The
                                     object is stored in the `value` attribute of the exception
        :raises KeyError: If the reference could not be resolved
        """
        if len(self.key) == 0:
            raise IndexError("List of keys is empty")
        # Find key index last (global) identifier-key in key list (from https://stackoverflow.com/a/6890255/10315508)
        try:
            last_identifier_index = next(i
                                         for i in reversed(range(len(self.key)))
                                         if self.key[i].get_identifier())
        except StopIteration:
            # If no identifier-key is contained in the list, we could try to resolve the path locally.
            # TODO implement local resolution
            raise NotImplementedError("We currently don't support local-only references without global identifier keys")

        resolved_keys: List[str] = []  # for more helpful error messages

        # First, resolve the identifier-key via the provider
        identifier: Identifier = self.key[last_identifier_index].get_identifier()  # type: ignore
        try:
            item: Referable = provider_.get_identifiable(identifier)
        except KeyError as e:
            raise KeyError("Could not resolve global reference key {}".format(identifier)) from e
        resolved_keys.append(str(identifier))

        # Now, follow path, given by remaining keys, recursively
        for key in self.key[last_identifier_index+1:]:
            if not isinstance(item, Namespace):
                raise TypeError("Object retrieved at {} is not a Namespace".format(" / ".join(resolved_keys)))
            try:
                item = item.get_referable(key.value)
            except KeyError as e:
                raise KeyError("Could not resolve id_short {} at {}".format(key.value, " / ".join(resolved_keys)))\
                    from e

        # Check type
        if not isinstance(item, self.type):
            raise UnexpectedTypeError(item, "Retrieved object {} is not an instance of referenced type {}"
                                            .format(item, self.type.__name__))
        return item

    def get_identifier(self) -> Identifier:
        """
        Retrieve the Identifier of the Identifiable object, which is referenced or in which the referenced Referable is
        contained.

        :raises ValueError: If this Reference does not include a Key with global KeyType (IRDI, IRI, CUSTOM)
        """
        try:
            last_identifier = next(key.get_identifier()
                                   for key in reversed(self.key)
                                   if key.get_identifier())
            return last_identifier  # type: ignore  # MyPy doesn't get the generator expression above
        except StopIteration:
            raise ValueError("Reference cannot be represented as an Identifier, since it does not contain a Key with "
                             "global KeyType (IRDI, IRI, CUSTOM)")

    def __repr__(self) -> str:
        return "AASReference(type={}, key={})".format(self.type.__name__, self.key)

    @staticmethod
    def from_referable(referable: Referable) -> "AASReference":
        """
        Construct a Reference to a given Referable AAS object

        This requires that the Referable object is Identifiable itself or is a child-, grand-child-, etc. object of an
        Identifiable object. Additionally, the object must be an instance of a known Referable type.

        :raises ValueError: If no Identifiable object is found while traversing the object's ancestors
        """
        # Get the first class from the base classes list (via inspect.getmro), that is contained in KEY_ELEMENTS_CLASSES
        from . import KEY_ELEMENTS_CLASSES
        try:
            ref_type = next(iter(t for t in inspect.getmro(type(referable)) if t in KEY_ELEMENTS_CLASSES))
        except StopIteration:
            ref_type = Referable

        ref: Referable = referable
        keys: List[Key] = []
        while True:
            keys.append(Key.from_referable(ref))
            if isinstance(ref, Identifiable):
                keys.reverse()
                return AASReference(tuple(keys), ref_type)
            if ref.parent is None or not isinstance(ref.parent, Referable):
                raise ValueError("The given Referable object is not embedded within an Identifiable object")
            ref = ref.parent


class Identifiable(Referable, metaclass=abc.ABCMeta):
    """
    An element that has a globally unique identifier.

    << abstract >>

    :ivar administration: Administrative information of an identifiable element.
    :ivar identification: The globally unique identification of the element.
    """
    @abc.abstractmethod
    def __init__(self):
        super().__init__()
        self.administration: Optional[AdministrativeInformation] = None
        self.identification: Identifier = Identifier("", IdentifierType.IRDI)

    def __repr__(self) -> str:
        return "{}[{}]".format(self.__class__.__name__, self.identification)


class HasSemantics(metaclass=abc.ABCMeta):
    """
    Element that can have a semantic definition.

    << abstract >>

    :ivar semantic_id: Identifier of the semantic definition of the element. It is called semantic id of the element.
                       The semantic id may either reference an external global id or it may reference a referable model
                       element of kind=Type that defines the semantics of the element.
    """
    @abc.abstractmethod
    def __init__(self):
        super().__init__()
        self.semantic_id: Optional[Reference] = None


class HasKind(metaclass=abc.ABCMeta):
    """
    An element with a kind is an element that can either represent a type or an instance.
    Default for an element is that it is representing an instance.

    << abstract >>

    :ivar kind: Kind of the element: either type or instance. Default = Instance.
    """
    @abc.abstractmethod
    def __init__(self):
        super().__init__()
        self._kind: ModelingKind = ModelingKind.INSTANCE

    @property
    def kind(self):
        return self._kind


class Constraint(metaclass=abc.ABCMeta):
    """
    A constraint is used to further qualify an element.

    << abstract >>
    """
    @abc.abstractmethod
    def __init__(self):
        pass


class Qualifiable(metaclass=abc.ABCMeta):
    """
    The value of a qualifiable element may be further qualified by one or more qualifiers or complex formulas.

    << abstract >>

    :ivar qualifier: Unordered list of Constraints that gives additional qualification of a qualifiable element.
    """
    @abc.abstractmethod
    def __init__(self):
        super().__init__()
        self.qualifier: Set[Constraint] = set()


class Formula(Constraint):
    """
    A formula is used to describe constraints by a logical expression.

    :ivar depends_on: Unordered list of references to referable or even external global elements that are used in the
                      logical expression. The value of the referenced elements needs to be accessible so that
                      it can be evaluated in the formula to true or false in the corresponding logical expression
                      it is used in.
    """

    def __init__(self,
                 depends_on: Optional[Set[Reference]] = None):
        """
        Initializer of Formula

        :param depends_on: Unordered of references to referable or even external global elements that are used in the
                           logical expression. The value of the referenced elements needs to be accessible so that
                           it can be evaluated in the formula to true or false in the corresponding logical expression
                           it is used in.

        TODO: Add instruction what to do after construction
        """
        super().__init__()
        self.depends_on: Set[Reference] = set() if depends_on is None else depends_on


class Qualifier(Constraint, HasSemantics):
    """
    A qualifier is a type-value pair that makes additional statements w.r.t. the value of the element.

    :ivar type: The type of the qualifier that is applied to the element.
    :ivar value_type: Data type of the qualifier value
    :ivar value: The value of the qualifier.
                 Constraint AASd-006: if both, the value and the valueId are present then the value needs to be
                                      identical to the value of the referenced coded value in Qualifier/valueId.
    :ivar value_id: Reference to the global unique id of a coded value.
    :ivar semantic_id: The semantic_id defined in the HasSemantics class.
    """

    def __init__(self,
                 type_: QualifierType,
                 value_type: DataTypeDef,
                 value: Optional[ValueDataType] = None,
                 value_id: Optional[Reference] = None,
                 semantic_id: Optional[Reference] = None):
        """
        Initializer of Qualifier

        :param type_: The type of the qualifier that is applied to the element.
        :param value_type: Data type of the qualifier value
        :param value: The value of the qualifier.
        :param value_id: Reference to the global unique id of a coded value.
        :param semantic_id: The semantic_id defined in the HasSemantics class.

        TODO: Add instruction what to do after construction
        """
        super().__init__()
        self.type: QualifierType = type_
        self.value_type: Type[datatypes.AnyXSDType] = value_type
        self._value: Optional[ValueDataType] = datatypes.trivial_cast(value, value_type) if value is not None else None
        self.value_id: Optional[Reference] = value_id
        self.semantic_id: Optional[Reference] = semantic_id

    def __repr__(self) -> str:
        return "Qualifier(type={})".format(self.type)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value) -> None:
        if value is None:
            self._value = None
        else:
            self._value = datatypes.trivial_cast(value, self.value_type)


class ValueReferencePair:
    """
    A value reference pair within a value list. Each value has a global unique id defining its semantic.

    << Data Type >>

    :ivar value: The value of the referenced concept definition of the value in value_id
    :ivar value_id: Global unique id of the value.
    """

    def __init__(self,
                 value_type: DataTypeDef,
                 value: ValueDataType,
                 value_id: Reference):
        """
        Initializer of ValueReferencePair

        :param value: The value of the referenced concept definition of the value in value_id
        :param value_id: Global unique id of the value.
        :param value_type: XSD datatype of the value (this is not compliant to the DotAAS meta model)

        TODO: Add instruction what to do after construction
        """
        self.value_type: Type[datatypes.AnyXSDType] = value_type
        self.value_id: Reference = value_id
        self._value: ValueDataType = datatypes.trivial_cast(value, value_type)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value) -> None:
        if value is None:
            raise AttributeError('Value can not be None')
        else:
            self._value = datatypes.trivial_cast(value, self.value_type)

    def __repr__(self) -> str:
        return "ValueReferencePair(value_type={}, value={}, value_id={})".format(self.value_type,
                                                                                 self.value,
                                                                                 self.value_id)


ValueList = Set[ValueReferencePair]


class Namespace(metaclass=abc.ABCMeta):
    """
    Abstract baseclass for all objects which form a Namespace to hold Referable objects and resolve them by their
    id_short.

    A Namespace can contain multiple NamespaceSets, which contain Referable objects of different types. However, the
    id_short of each object must be unique across all NamespaceSets of one Namespace.

    :ivar namespace_element_sets: A list of all NamespaceSets of this Namespace
    """
    @abc.abstractmethod
    def __init__(self) -> None:
        super().__init__()
        self.namespace_element_sets: List[NamespaceSet] = []

    def get_referable(self, id_short: str) -> Referable:
        """
        Find a Referable in this Namespaces by its id_short

        :raises KeyError: If no such Referable can be found
        """
        for dict_ in self.namespace_element_sets:
            if id_short in dict_:
                return dict_.get_referable(id_short)
        raise KeyError("Referable with id_short {} not found in this namespace".format(id_short))

    def remove_referable(self, id_short: str) -> None:
        """
        Remove a Referable from this Namespace by its id_short

        :raises KeyError: If no such Referable can be found
        """
        for dict_ in self.namespace_element_sets:
            if id_short in dict_:
                return dict_.remove(id_short)
        raise KeyError("Referable with id_short {} not found in this namespace".format(id_short))

    def __iter__(self) -> Iterator[_RT]:
        return itertools.chain.from_iterable(self.namespace_element_sets)


class NamespaceSet(MutableSet[_RT], Generic[_RT]):
    """
    Helper class for storing Referable objects of a given type in a Namespace and find them by their id_short.

    This class behaves much like a set of Referable objects of a defined type, but uses a dict internally to rapidly
    find those objects by their id_short. Additionally, it manages the `parent` attribute of the stored Referables and
    ensures the uniqueness of their id_short within the Namespace.

    Use `add()`, `remove()`, `pop()`, `discard()`, `clear()`, `len()`, `x in` checks and iteration  just like on a
    normal set of Referables. To get a referable by its id_short, use `get_referable()` or `get()` (the latter one
    allows a default argument and returns None instead of raising a KeyError). As a bonus, the `x in` check supports
    checking for existence of id_short *or* a concrete Referable object.
    """
    def __init__(self, parent: Namespace, items: Iterable[_RT] = ()) -> None:
        """
        Initialize a new NamespaceSet.

        This initializer automatically takes care of adding this set to the `namespace_element_sets` list of the
        Namespace.

        :param parent: The Namespace this set belongs to
        :param items: A given list of Referable items to be added to the set
        :raises KeyError: When `items` contains multiple objects with same id_short
        """
        self.parent = parent
        parent.namespace_element_sets.append(self)
        self._backend: Dict[str, _RT] = {}
        try:
            for i in items:
                self.add(i)
        except Exception:
            # Do a rollback, when an exception occurs while adding items
            self.clear()
            raise

    def __contains__(self, x: object) -> bool:
        if isinstance(x, str):
            return x in self._backend
        elif isinstance(x, Referable):
            return self._backend.get(x.id_short) is x
        else:
            return False

    def __len__(self) -> int:
        return len(self._backend)

    def __iter__(self) -> Iterator[_RT]:
        return iter(self._backend.values())

    def add(self, value: _RT):
        for set_ in self.parent.namespace_element_sets:
            if value.id_short in set_:
                raise KeyError("Referable with id_short '{}' is already present in {}"
                               .format(value.id_short,
                                       "this set of objects"
                                       if set_ is self else "another set in the same namespace"))
        if value.parent is not None and value.parent is not self.parent:
            raise ValueError("Object has already a parent, but it must not be part of two namespaces.")
            # TODO remove from current parent instead (allow moving)?
        value.parent = self.parent
        self._backend[value.id_short] = value

    def remove(self, item: Union[str, _RT]):
        if isinstance(item, str):
            del self._backend[item]
        else:
            item_in_dict = self._backend[item.id_short]
            if item_in_dict is not item:
                raise KeyError("Item not found in NamespaceDict (other item with same id_short exists)")
            item.parent = None
            del self._backend[item.id_short]

    def discard(self, x: _RT) -> None:
        if x not in self:
            return
        self.remove(x)

    def pop(self) -> _RT:
        _, value = self._backend.popitem()
        value.parent = None
        return value

    def clear(self) -> None:
        for value in self._backend.values():
            value.parent = None
        self._backend.clear()

    def get_referable(self, key) -> _RT:
        """
        Find an object in this set by its id_short

        :raises KeyError: If no such object can be found
        """
        return self._backend[key]

    def get(self, key, default: Optional[_RT] = None) -> Optional[_RT]:
        """
        Find an object in this set by its id_short, with fallback parameter

        :param default: An object to be returned, if no object with the given id_short is found
        :return: The Referable object with the given id_short in the set. Otherwise the `default` object or None, if
                 none is given.
        """
        return self._backend.get(key, default)

    def update_nss_from(self, other: "NamespaceSet"):
        """
        Update a NamespaceSet from a given NamespaceSet.

        WARNING: By updating, the "other" NamespaceSet gets destroyed.

        :param other: The NamespaceSet to update from
        """
        referables_to_add: List[Referable] = []  # objects from the other nss to add to self
        referables_to_remove: List[Referable] = []  # objects to remove from self
        for other_referable in other:
            try:
                referable = self._backend[other_referable.id_short]
                if type(referable) is type(other_referable):
                    # referable is the same as other referable
                    referable.update_from(other_referable, update_source=True)
            except KeyError:
                # other referable is not in NamespaceSet
                referables_to_add.append(other_referable)
        for id_short, referable in self._backend.items():
            if not other.get(id_short):
                # referable does not exist in the other NamespaceSet
                referables_to_remove.append(referable)
        for referable_to_add in referables_to_add:
            other.remove(referable_to_add)
            self.add(referable_to_add)  # type: ignore
        for referable_to_remove in referables_to_remove:
            self.remove(referable_to_remove)  # type: ignore


class OrderedNamespaceSet(NamespaceSet[_RT], MutableSequence[_RT], Generic[_RT]):
    """
    A specialized version of NamespaceSet, that keeps track of the order of the stored Referable objects.

    Additionally to the MutableSet interface of NamespaceSet, this class provides a set-like interface (actually it
    is derived from MutableSequence). However, we don't permit duplicate entries in the ordered list of objects.
    """
    def __init__(self, parent: Namespace, items: Iterable[_RT] = ()) -> None:
        """
        Initialize a new OrderedNamespaceSet.

        This initializer automatically takes care of adding this set to the `namespace_element_sets` list of the
        Namespace.

        :param parent: The Namespace this set belongs to
        :param items: A given list of Referable items to be added to the set
        :raises KeyError: When `items` contains multiple objects with same id_short
        """
        self._order: List[_RT] = []
        super().__init__(parent, items)

    def __iter__(self) -> Iterator[_RT]:
        return iter(self._order)

    def add(self, value: _RT):
        super().add(value)
        self._order.append(value)

    def remove(self, item: Union[str, _RT]):
        if isinstance(item, str):
            item = self.get_referable(item)
        super().remove(item)
        self._order.remove(item)

    def pop(self, i: Optional[int] = None):
        if i is None:
            value = super().pop()
            self._order.remove(value)
        else:
            value = self._order.pop(i)
            super().remove(value)
        return value

    def clear(self) -> None:
        super().clear()
        self._order.clear()

    def insert(self, index: int, object_: _RT) -> None:
        super().add(object_)
        self._order.insert(index, object_)

    @overload
    def __getitem__(self, i: int) -> _RT: ...

    @overload
    def __getitem__(self, s: slice) -> MutableSequence[_RT]: ...

    def __getitem__(self, s: Union[int, slice]) -> Union[_RT, MutableSequence[_RT]]:
        return self._order[s]

    @overload
    def __setitem__(self, i: int, o: _RT) -> None: ...

    @overload
    def __setitem__(self, s: slice, o: Iterable[_RT]) -> None: ...

    def __setitem__(self, s, o) -> None:
        if isinstance(s, int):
            deleted_items = [self._order[s]]
            super().add(o)
            self._order[s] = o
        else:
            deleted_items = self._order[s]
            new_items = itertools.islice(o, len(deleted_items))
            successful_new_items = []
            try:
                for i in new_items:
                    super().add(i)
                    successful_new_items.append(i)
            except Exception:
                # Do a rollback, when an exception occurs while adding items
                for i in successful_new_items:
                    super().remove(i)
                raise
            self._order[s] = new_items
        for i in deleted_items:
            super().remove(i)

    @overload
    def __delitem__(self, i: int) -> None: ...

    @overload
    def __delitem__(self, i: slice) -> None: ...

    def __delitem__(self, i: Union[int, slice]) -> None:
        if isinstance(i, int):
            i = slice(i, i+1)
        for o in self._order[i]:
            super().remove(o)
        del self._order[i]



# Copyright (c) 2020 PyI40AAS Contributors
#
# This program and the accompanying materials are made available under the terms of the Eclipse Public License v. 2.0
# which is available at https://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0 which is available
# at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
"""
Module for serializing Asset Administration Shell objects to the official JSON format

The module provides an custom JSONEncoder class `AASToJsonEncoder` to be used with the Python standard `json` module.
It contains a custom `default` function which converts PyI40AAS objects to simple python types for an automatic
JSON serialization. Additionally, there's the `write_aas_json_file()` function, that takes a complete ObjectStore and
writes all contained AAS objects into a JSON file.

This job is performed in an iterative approach: The `default()` function gets called for every object and checks if an
object is an PyI40AAS object. In this case, it calls a special function for the respective PyI40AAS class which converts
the object (but not the contained objects) into a simple Python dict, which is serializable. Any contained
PyI40AAS objects are included into the dict as they are to be converted later on. The special helper function
`abstract_classes_to_json()` is called by most of the conversion functions to handle all the attributes of abstract base
classes.
"""
import base64
import inspect
from typing import List, Dict, IO, Optional, Type
import json

from aas import model
from .. import _generic


class AASToJsonEncoder(json.JSONEncoder):
    """
    Custom JSONDecoder class to use the `json` module for serializing Asset Administration Shell data into the
    official JSON format

    The class overrides the `default()` method to transform PyI40AAS objects into dicts that may be serialized by the
    standard encode method. Typical usage:

        json_string = json.dumps(data, cls=AASToJsonEncoder)

    :cvar stripped: If True, the JSON objects will be serialized in a stripped manner, excluding some attributes.
                    Defaults to False.
                    See https://git.rwth-aachen.de/acplt/pyi40aas/-/issues/91
    """
    stripped = False

    def default(self, obj: object) -> object:
        if isinstance(obj, model.AssetAdministrationShell):
            return self._asset_administration_shell_to_json(obj)
        if isinstance(obj, model.Identifier):
            return self._identifier_to_json(obj)
        if isinstance(obj, model.AdministrativeInformation):
            return self._administrative_information_to_json(obj)
        if isinstance(obj, model.Reference):
            return self._reference_to_json(obj)
        if isinstance(obj, model.Key):
            return self._key_to_json(obj)
        if isinstance(obj, model.ValueReferencePair):
            return self._value_reference_pair_to_json(obj)
        if isinstance(obj, model.Asset):
            return self._asset_to_json(obj)
        if isinstance(obj, model.Submodel):
            return self._submodel_to_json(obj)
        if isinstance(obj, model.Operation):
            return self._operation_to_json(obj)
        if isinstance(obj, model.OperationVariable):
            return self._operation_variable_to_json(obj)
        if isinstance(obj, model.Capability):
            return self._capability_to_json(obj)
        if isinstance(obj, model.BasicEvent):
            return self._basic_event_to_json(obj)
        if isinstance(obj, model.Entity):
            return self._entity_to_json(obj)
        if isinstance(obj, model.View):
            return self._view_to_json(obj)
        if isinstance(obj, model.ConceptDictionary):
            return self._concept_dictionary_to_json(obj)
        if isinstance(obj, model.ConceptDescription):
            return self._concept_description_to_json(obj)
        if isinstance(obj, model.Property):
            return self._property_to_json(obj)
        if isinstance(obj, model.Range):
            return self._range_to_json(obj)
        if isinstance(obj, model.MultiLanguageProperty):
            return self._multi_language_property_to_json(obj)
        if isinstance(obj, model.File):
            return self._file_to_json(obj)
        if isinstance(obj, model.Blob):
            return self._blob_to_json(obj)
        if isinstance(obj, model.ReferenceElement):
            return self._reference_element_to_json(obj)
        if isinstance(obj, model.SubmodelElementCollection):
            return self._submodel_element_collection_to_json(obj)
        if isinstance(obj, model.AnnotatedRelationshipElement):
            return self._annotated_relationship_element_to_json(obj)
        if isinstance(obj, model.RelationshipElement):
            return self._relationship_element_to_json(obj)
        if isinstance(obj, model.Qualifier):
            return self._qualifier_to_json(obj)
        if isinstance(obj, model.Formula):
            return self._formula_to_json(obj)
        return super().default(obj)

    @classmethod
    def _abstract_classes_to_json(cls, obj: object) -> Dict[str, object]:
        """
        transformation function to serialize abstract classes from model.base which are inherited by many classes

        :param obj: object which must be serialized
        :return: dict with the serialized attributes of the abstract classes this object inherits from
        """
        data = {}
        if isinstance(obj, model.Referable):
            data['idShort'] = obj.id_short
            if obj.category:
                data['category'] = obj.category
            if obj.description:
                data['description'] = cls._lang_string_set_to_json(obj.description)
            try:
                ref_type = next(iter(t for t in inspect.getmro(type(obj)) if t in model.KEY_ELEMENTS_CLASSES))
            except StopIteration as e:
                raise TypeError("Object of type {} is Referable but does not inherit from a known AAS type"
                                .format(obj.__class__.__name__)) from e
            data['modelType'] = {'name': ref_type.__name__}
        if isinstance(obj, model.Identifiable):
            data['identification'] = obj.identification
            if obj.administration:
                data['administration'] = obj.administration
        if isinstance(obj, model.HasSemantics):
            if obj.semantic_id:
                data['semanticId'] = obj.semantic_id
        if isinstance(obj, model.HasKind):
            if obj.kind is model.ModelingKind.TEMPLATE:
                data['kind'] = _generic.MODELING_KIND[obj.kind]
        if isinstance(obj, model.Qualifiable) and not cls.stripped:
            if obj.qualifier:
                data['qualifiers'] = list(obj.qualifier)
        return data

    # #############################################################
    # transformation functions to serialize classes from model.base
    # #############################################################

    @classmethod
    def _lang_string_set_to_json(cls, obj: model.LangStringSet) -> List[Dict[str, object]]:
        return [{'language': k, 'text': v}
                for k, v in obj.items()]

    @classmethod
    def _key_to_json(cls, obj: model.Key) -> Dict[str, object]:
        """
        serialization of an object from class Key to json

        :param obj: object of class Key
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update({'type': _generic.KEY_ELEMENTS[obj.type],
                     'idType': _generic.KEY_TYPES[obj.id_type],
                     'value': obj.value,
                     'local': obj.local})
        return data

    @classmethod
    def _administrative_information_to_json(cls, obj: model.AdministrativeInformation) -> Dict[str, object]:
        """
        serialization of an object from class AdministrativeInformation to json

        :param obj: object of class AdministrativeInformation
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if obj.version:
            data['version'] = obj.version
            if obj.revision:
                data['revision'] = obj.revision
        return data

    @classmethod
    def _identifier_to_json(cls, obj: model.Identifier) -> Dict[str, object]:
        """
        serialization of an object from class Identifier to json

        :param obj: object of class Identifier
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data['id'] = obj.id
        data['idType'] = _generic.IDENTIFIER_TYPES[obj.id_type]
        return data

    @classmethod
    def _reference_to_json(cls, obj: model.Reference) -> Dict[str, object]:
        """
        serialization of an object from class Reference to json

        :param obj: object of class Reference
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data['keys'] = list(obj.key)
        return data

    @classmethod
    def _constraint_to_json(cls, obj: model.Constraint) -> Dict[str, object]:  # TODO check if correct for each class
        """
        serialization of an object from class Constraint to json

        :param obj: object of class Constraint
        :return: dict with the serialized attributes of this object
        """
        CONSTRAINT_CLASSES = [model.Qualifier, model.Formula]
        try:
            const_type = next(iter(t for t in inspect.getmro(type(obj)) if t in CONSTRAINT_CLASSES))
        except StopIteration as e:
            raise TypeError("Object of type {} is a Constraint but does not inherit from a known AAS Constraint type"
                            .format(obj.__class__.__name__)) from e
        return {'modelType': {'name': const_type.__name__}}

    @classmethod
    def _namespace_to_json(cls, obj):  # not in specification yet
        """
        serialization of an object from class Namespace to json

        :param obj: object of class Namespace
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        return data

    @classmethod
    def _formula_to_json(cls, obj: model.Formula) -> Dict[str, object]:
        """
        serialization of an object from class Formula to json

        :param obj: object of class Formula
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update(cls._constraint_to_json(obj))
        if obj.depends_on:
            data['dependsOn'] = list(obj.depends_on)
        return data

    @classmethod
    def _qualifier_to_json(cls, obj: model.Qualifier) -> Dict[str, object]:
        """
        serialization of an object from class Qualifier to json

        :param obj: object of class Qualifier
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update(cls._constraint_to_json(obj))
        if obj.value:
            data['value'] = model.datatypes.xsd_repr(obj.value) if obj.value is not None else None
        if obj.value_id:
            data['valueId'] = obj.value_id
        data['valueType'] = model.datatypes.XSD_TYPE_NAMES[obj.value_type]
        data['type'] = obj.type
        return data

    @classmethod
    def _value_reference_pair_to_json(cls, obj: model.ValueReferencePair) -> Dict[str, object]:
        """
        serialization of an object from class ValueReferencePair to json

        :param obj: object of class ValueReferencePair
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update({'value': model.datatypes.xsd_repr(obj.value),
                     'valueId': obj.value_id,
                     'valueType': model.datatypes.XSD_TYPE_NAMES[obj.value_type]})
        return data

    @classmethod
    def _value_list_to_json(cls, obj: model.ValueList) -> Dict[str, object]:
        """
        serialization of an object from class ValueList to json

        :param obj: object of class ValueList
        :return: dict with the serialized attributes of this object
        """
        return {'valueReferencePairTypes': list(obj)}

    # ############################################################
    # transformation functions to serialize classes from model.aas
    # ############################################################

    @classmethod
    def _view_to_json(cls, obj: model.View) -> Dict[str, object]:
        """
        serialization of an object from class View to json

        :param obj: object of class View
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if obj.contained_element:
            data['containedElements'] = list(obj.contained_element)
        return data

    @classmethod
    def _asset_to_json(cls, obj: model.Asset) -> Dict[str, object]:
        """
        serialization of an object from class Asset to json

        :param obj: object of class Asset
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data['kind'] = _generic.ASSET_KIND[obj.kind]
        if obj.asset_identification_model:
            data['assetIdentificationModel'] = obj.asset_identification_model
        if obj.bill_of_material:
            data['billOfMaterial'] = obj.bill_of_material
        return data

    @classmethod
    def _concept_description_to_json(cls, obj: model.ConceptDescription) -> Dict[str, object]:
        """
        serialization of an object from class ConceptDescription to json

        :param obj: object of class ConceptDescription
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if obj.is_case_of:
            data['isCaseOf'] = list(obj.is_case_of)

        if isinstance(obj, model.concept.IEC61360ConceptDescription):
            cls._append_iec61360_concept_description_attrs(obj, data)

        return data

    @classmethod
    def _append_iec61360_concept_description_attrs(cls, obj: model.concept.IEC61360ConceptDescription,
                                                   data: Dict[str, object]) -> None:
        """
        Add the 'embeddedDataSpecifications' attribute to IEC61360ConceptDescription's JSON representation.

        `IEC61360ConceptDescription` is not a distinct class according DotAAS, but instead is built by referencing
        "DataSpecificationIEC61360" as dataSpecification. However, we implemented it as an explicit class, inheriting
        from ConceptDescription, but we want to generate compliant JSON documents. So, we fake the JSON structure of an
        object with dataSpecifications.
        """
        data_spec: Dict[str, object] = {
            'preferredName': cls._lang_string_set_to_json(obj.preferred_name)
        }
        if obj.data_type is not None:
            data_spec['dataType'] = _generic.IEC61360_DATA_TYPES[obj.data_type]
        if obj.definition is not None:
            data_spec['definition'] = cls._lang_string_set_to_json(obj.definition)
        if obj.short_name is not None:
            data_spec['shortName'] = cls._lang_string_set_to_json(obj.short_name)
        if obj.unit is not None:
            data_spec['unit'] = obj.unit
        if obj.unit_id is not None:
            data_spec['unitId'] = obj.unit_id
        if obj.source_of_definition is not None:
            data_spec['sourceOfDefinition'] = obj.source_of_definition
        if obj.symbol is not None:
            data_spec['symbol'] = obj.symbol
        if obj.value_format is not None:
            data_spec['valueFormat'] = model.datatypes.XSD_TYPE_NAMES[obj.value_format]
        if obj.value_list is not None:
            data_spec['valueList'] = cls._value_list_to_json(obj.value_list)
        if obj.value is not None:
            data_spec['value'] = model.datatypes.xsd_repr(obj.value) if obj.value is not None else None
        if obj.value_id is not None:
            data_spec['valueId'] = obj.value_id
        if obj.level_types:
            data_spec['levelType'] = [_generic.IEC61360_LEVEL_TYPES[lt] for lt in obj.level_types]
        data['embeddedDataSpecifications'] = [
            {'dataSpecification': model.Reference((
                model.Key(model.KeyElements.GLOBAL_REFERENCE, False,
                          "http://admin-shell.io/DataSpecificationTemplates/DataSpecificationIEC61360/2/0",
                          model.KeyType.IRI),)),
             'dataSpecificationContent': data_spec}
        ]

    @classmethod
    def _concept_dictionary_to_json(cls, obj: model.ConceptDictionary) -> Dict[str, object]:
        """
        serialization of an object from class ConceptDictionary to json

        :param obj: object of class ConceptDictionary
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if obj.concept_description:
            data['conceptDescriptions'] = list(obj.concept_description)
        return data

    @classmethod
    def _asset_administration_shell_to_json(cls, obj: model.AssetAdministrationShell) -> Dict[str, object]:
        """
        serialization of an object from class AssetAdministrationShell to json

        :param obj: object of class AssetAdministrationShell
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update(cls._namespace_to_json(obj))
        if obj.derived_from:
            data["derivedFrom"] = obj.derived_from
        data["asset"] = obj.asset
        if not cls.stripped and obj.submodel:
            data["submodels"] = list(obj.submodel)
        if not cls.stripped and obj.view:
            data["views"] = list(obj.view)
        if obj.concept_dictionary:
            data["conceptDictionaries"] = list(obj.concept_dictionary)
        if obj.security:
            data["security"] = obj.security
        return data

    # #################################################################
    # transformation functions to serialize classes from model.security
    # #################################################################

    @classmethod
    def _security_to_json(cls, obj: model.Security) -> Dict[str, object]:  # has no attributes in our implementation
        """
        serialization of an object from class Security to json

        :param obj: object of class Security
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        return data

    # #################################################################
    # transformation functions to serialize classes from model.submodel
    # #################################################################

    @classmethod
    def _submodel_to_json(cls, obj: model.Submodel) -> Dict[str, object]:  # TODO make kind optional
        """
        serialization of an object from class Submodel to json

        :param obj: object of class Submodel
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if not cls.stripped and obj.submodel_element != set():
            data['submodelElements'] = list(obj.submodel_element)
        return data

    @classmethod
    def _data_element_to_json(cls, obj: model.DataElement) -> Dict[str, object]:  # no attributes in specification yet
        """
        serialization of an object from class DataElement to json

        :param obj: object of class DataElement
        :return: dict with the serialized attributes of this object
        """
        return {}

    @classmethod
    def _property_to_json(cls, obj: model.Property) -> Dict[str, object]:
        """
        serialization of an object from class Property to json

        :param obj: object of class Property
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data['value'] = model.datatypes.xsd_repr(obj.value) if obj.value is not None else None
        if obj.value_id:
            data['valueId'] = obj.value_id
        data['valueType'] = model.datatypes.XSD_TYPE_NAMES[obj.value_type]
        return data

    @classmethod
    def _multi_language_property_to_json(cls, obj: model.MultiLanguageProperty) -> Dict[str, object]:
        """
        serialization of an object from class MultiLanguageProperty to json

        :param obj: object of class MultiLanguageProperty
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if obj.value:
            data['value'] = cls._lang_string_set_to_json(obj.value)
        if obj.value_id:
            data['valueId'] = obj.value_id
        return data

    @classmethod
    def _range_to_json(cls, obj: model.Range) -> Dict[str, object]:
        """
        serialization of an object from class Range to json

        :param obj: object of class Range
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update({'valueType': model.datatypes.XSD_TYPE_NAMES[obj.value_type],
                     'min': model.datatypes.xsd_repr(obj.min) if obj.min is not None else None,
                     'max': model.datatypes.xsd_repr(obj.max) if obj.max is not None else None})
        return data

    @classmethod
    def _blob_to_json(cls, obj: model.Blob) -> Dict[str, object]:
        """
        serialization of an object from class Blob to json

        :param obj: object of class Blob
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data['mimeType'] = obj.mime_type
        if obj.value is not None:
            data['value'] = base64.b64encode(obj.value).decode()
        return data

    @classmethod
    def _file_to_json(cls, obj: model.File) -> Dict[str, object]:
        """
        serialization of an object from class File to json

        :param obj: object of class File
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update({'value': obj.value, 'mimeType': obj.mime_type})
        return data

    @classmethod
    def _reference_element_to_json(cls, obj: model.ReferenceElement) -> Dict[str, object]:
        """
        serialization of an object from class Reference to json

        :param obj: object of class Reference
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if obj.value:
            data['value'] = obj.value
        return data

    @classmethod
    def _submodel_element_collection_to_json(cls, obj: model.SubmodelElementCollection) -> Dict[str, object]:
        """
        serialization of an object from class SubmodelElementCollectionOrdered and SubmodelElementCollectionUnordered to
        json

        :param obj: object of class SubmodelElementCollectionOrdered and SubmodelElementCollectionUnordered
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if not cls.stripped and obj.value:
            data['value'] = list(obj.value)
        data['ordered'] = obj.ordered
        return data

    @classmethod
    def _relationship_element_to_json(cls, obj: model.RelationshipElement) -> Dict[str, object]:
        """
        serialization of an object from class RelationshipElement to json

        :param obj: object of class RelationshipElement
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update({'first': obj.first, 'second': obj.second})
        return data

    @classmethod
    def _annotated_relationship_element_to_json(cls, obj: model.AnnotatedRelationshipElement) -> Dict[str, object]:
        """
        serialization of an object from class AnnotatedRelationshipElement to json

        :param obj: object of class AnnotatedRelationshipElement
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data.update({'first': obj.first, 'second': obj.second})
        if not cls.stripped and obj.annotation:
            data['annotation'] = list(obj.annotation)
        return data

    @classmethod
    def _operation_variable_to_json(cls, obj: model.OperationVariable) -> Dict[str, object]:
        """
        serialization of an object from class OperationVariable to json

        :param obj: object of class OperationVariable
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data['value'] = obj.value
        return data

    @classmethod
    def _operation_to_json(cls, obj: model.Operation) -> Dict[str, object]:
        """
        serialization of an object from class Operation to json

        :param obj: object of class Operation
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if obj.input_variable:
            data['inputVariable'] = list(obj.input_variable)
        if obj.output_variable:
            data['outputVariable'] = list(obj.output_variable)
        if obj.in_output_variable:
            data['inoutputVariable'] = list(obj.in_output_variable)
        return data

    @classmethod
    def _capability_to_json(cls, obj: model.Capability) -> Dict[str, object]:
        """
        serialization of an object from class Capability to json

        :param obj: object of class Capability
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        # no attributes in specification yet
        return data

    @classmethod
    def _entity_to_json(cls, obj: model.Entity) -> Dict[str, object]:
        """
        serialization of an object from class Entity to json

        :param obj: object of class Entity
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        if not cls.stripped and obj.statement:
            data['statements'] = list(obj.statement)
        data['entityType'] = _generic.ENTITY_TYPES[obj.entity_type]
        if obj.asset:
            data['asset'] = obj.asset
        return data

    @classmethod
    def _event_to_json(cls, obj: model.Event) -> Dict[str, object]:  # no attributes in specification yet
        """
        serialization of an object from class Event to json

        :param obj: object of class Event
        :return: dict with the serialized attributes of this object
        """
        return {}

    @classmethod
    def _basic_event_to_json(cls, obj: model.BasicEvent) -> Dict[str, object]:
        """
        serialization of an object from class BasicEvent to json

        :param obj: object of class BasicEvent
        :return: dict with the serialized attributes of this object
        """
        data = cls._abstract_classes_to_json(obj)
        data['observed'] = obj.observed
        return data


class StrippedAASToJsonEncoder(AASToJsonEncoder):
    """
    AASToJsonEncoder for stripped objects. Used in the HTTP API.
    See https://git.rwth-aachen.de/acplt/pyi40aas/-/issues/91
    """
    stripped = True


def _select_encoder(stripped: bool, encoder: Optional[Type[AASToJsonEncoder]] = None) -> Type[AASToJsonEncoder]:
    """
    Returns the correct encoder based on the stripped parameter. If an encoder class is given, stripped is ignored.

    :param stripped: If true, an encoder for parsing stripped JSON objects is selected. Ignored if an encoder class is
                     specified.
    :param encoder: Is returned, if specified.
    :return: A AASToJsonEncoder (sub)class.
    """
    if encoder is not None:
        return encoder
    return AASToJsonEncoder if not stripped else StrippedAASToJsonEncoder


def _create_dict(data: model.AbstractObjectStore) -> dict:
    # separate different kind of objects
    assets = []
    asset_administration_shells = []
    submodels = []
    concept_descriptions = []
    for obj in data:
        if isinstance(obj, model.Asset):
            assets.append(obj)
        if isinstance(obj, model.AssetAdministrationShell):
            asset_administration_shells.append(obj)
        if isinstance(obj, model.Submodel):
            submodels.append(obj)
        if isinstance(obj, model.ConceptDescription):
            concept_descriptions.append(obj)
    dict_ = {
        'assetAdministrationShells': asset_administration_shells,
        'submodels': submodels,
        'assets': assets,
        'conceptDescriptions': concept_descriptions,
    }
    return dict_


def object_store_to_json(data: model.AbstractObjectStore, stripped: bool = False,
                         encoder: Optional[Type[AASToJsonEncoder]] = None, **kwargs) -> str:
    """
    Create a json serialization of a set of AAS objects according to 'Details of the Asset Administration Shell',
    chapter 5.5

    :param data: ObjectStore which contains different objects of the AAS meta model which should be serialized to a
                 JSON file
    :param stripped: If true, objects are serialized to stripped json objects..
                     See https://git.rwth-aachen.de/acplt/pyi40aas/-/issues/91
                     This parameter is ignored if an encoder class is specified.
    :param encoder: The encoder class used to encoder the JSON objects
    :param kwargs: Additional keyword arguments to be passed to json.dump()
    """
    encoder_ = _select_encoder(stripped, encoder)
    # serialize object to json
    return json.dumps(_create_dict(data), cls=encoder_, **kwargs)


def write_aas_json_file(file: IO, data: model.AbstractObjectStore, stripped: bool = False,
                        encoder: Optional[Type[AASToJsonEncoder]] = None, **kwargs) -> None:
    """
    Write a set of AAS objects to an Asset Administration Shell JSON file according to 'Details of the Asset
    Administration Shell', chapter 5.5

    :param file: A file-like object to write the JSON-serialized data to
    :param data: ObjectStore which contains different objects of the AAS meta model which should be serialized to a
                 JSON file
    :param stripped: If true, objects are serialized to stripped json objects..
                     See https://git.rwth-aachen.de/acplt/pyi40aas/-/issues/91
                     This parameter is ignored if an encoder class is specified.
    :param encoder: The encoder class used to encoder the JSON objects
    :param kwargs: Additional keyword arguments to be passed to json.dumps()
    """
    encoder_ = _select_encoder(stripped, encoder)
    # serialize object to json
    json.dump(_create_dict(data), file, cls=encoder_, **kwargs)
